# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""
Containers for PyROS subproblem solve results.
"""

from pyomo.opt import check_optimal_termination


class ROSolveResults:
    """
    PyROS solver results object.

    Parameters
    ----------
    config : ConfigDict, optional
        User-specified solver settings.
    iterations : int, optional
        Number of iterations required.
    time : float, optional
        Total elapsed time (or wall time), in seconds.
    final_objective_value : float, optional
        Final objective function value to report.
    pyros_termination_condition : pyrosTerminationCondition, optional
        PyROS-specific termination condition.

    Attributes
    ----------
    config : ConfigDict
        User-specified solver settings.
    iterations : int
        Number of iterations required by PyROS.
    time : float
        Total elapsed time (or wall time), in seconds.
    final_objective_value : float
        Final objective function value to report.
        If a nominal objective focus was elected, then the
        value of the nominal objective function is reported.
        If a worst-case objective focus was elected, then
        the value of the worst-case objective function is reported.
    pyros_termination_condition : pyrosTerminationCondition
        Indicator of the manner of termination.
    """

    def __init__(
        self,
        config=None,
        iterations=None,
        time=None,
        final_objective_value=None,
        pyros_termination_condition=None,
    ):
        """Initialize self (see class docstring)."""
        self.config = config
        self.iterations = iterations
        self.time = time
        self.final_objective_value = final_objective_value
        self.pyros_termination_condition = pyros_termination_condition

    def __str__(self):
        """
        Generate string representation of self.
        Does not include any information about `self.config`.
        """
        lines = ["Termination stats:"]
        attr_name_format_dict = {
            "iterations": ("Iterations", "f'{val}'"),
            "time": ("Solve time (wall s)", "f'{val:.3f}'"),
            "final_objective_value": ("Final objective value", "f'{val:.4e}'"),
            "pyros_termination_condition": ("Termination condition", "f'{val}'"),
        }
        attr_desc_pad_length = max(
            len(desc) for desc, _ in attr_name_format_dict.values()
        )
        for attr_name, (attr_desc, fmt_str) in attr_name_format_dict.items():
            val = getattr(self, attr_name)
            val_str = eval(fmt_str) if val is not None else str(val)
            lines.append(f" {attr_desc:<{attr_desc_pad_length}s} : {val_str}")

        return "\n".join(lines)


class MasterResults:
    """
    Result of solving the master problem in a single PyROS iteration.

    Attributes
    ----------
    master_model : ConcreteModel
        Master model.
    feasibility_problem_results : SolverResults
        Feasibility problem subsolver results.
    master_results_list : list of SolverResults
        List of subsolver results for the master problem.
    pyros_termination_condition : None or pyrosTerminationCondition
        PyROS termination status established via solution of
        the master problem.
        If `None`, then no termination status has been established.
    """

    def __init__(
        self,
        master_model=None,
        feasibility_problem_results=None,
        master_results_list=None,
        pyros_termination_condition=None,
    ):
        """Initialize self (see class docstring)."""
        self.master_model = master_model
        self.feasibility_problem_results = feasibility_problem_results
        if master_results_list is None:
            self.master_results_list = []
        else:
            self.master_results_list = list(master_results_list)
        self.pyros_termination_condition = pyros_termination_condition

    @property
    def backup_solver_used(self):
        """
        bool : True if a backup solver was used for the master problem,
        False otherwise.
        """
        return len(self.master_results_list) > 1

    @property
    def feasibility_problem_success(self):
        """
        bool : True if the feasibility problem was solved
        successfully, False otherwise.
        """
        return self.feasibility_problem_results is None or check_optimal_termination(
            self.feasibility_problem_results
        )


class SeparationSolveCallResults:
    """
    Container for results of solve attempt for single separation
    problem.

    Parameters
    ----------
    solved_globally : bool
        True if separation problem was solved globally,
        False otherwise.
    results_list : list of pyomo.opt.results.SolverResults, optional
        Pyomo solver results for each subordinate optimizer invoked on
        the separation problem.
        For problems with non-discrete uncertainty set types,
        each entry corresponds to a single subordinate solver.
        For problems with discrete set types, the list may
        be empty (didn't need to use a subordinate solver to
        evaluate optimal separation solution), or the number
        of entries may be as high as the product of the number of
        subordinate local/global solvers provided (including backup)
        and the number of scenarios in the uncertainty set.
    scaled_violations : ComponentMap, optional
        Mapping from second-stage inequality constraints to floats equal
        to their scaled violations by separation problem solution
        stored in this result.
    violating_param_realization : list of float, optional
        Uncertain parameter realization for reported separation
        problem solution.
    auxiliary_param_values : list of float, optional
        Auxiliary parameter values corresponding to the
        uncertain parameter realization `violating_param_realization`.
    variable_values : ComponentMap, optional
        Second-stage DOF and state variable values for reported
        separation problem solution.
    found_violation : bool, optional
        True if violation of second-stage inequality constraint
        (i.e. constraint expression value) by reported separation
        solution was found to exceed tolerance, False otherwise.
    time_out : bool, optional
        True if PyROS time limit reached attempting to solve the
        separation problem, False otherwise.
    subsolver_error : bool, optional
        True if subsolvers found to be unable to solve separation
        problem of interest, False otherwise.
    discrete_set_scenario_index : None or int, optional
        If discrete set used to solve the problem, index of
        `violating_param_realization` as listed in the
        `scenarios` attribute of a ``DiscreteScenarioSet``
        instance. If discrete set not used, pass None.
    backup_solver_used : bool, optional
        True if a backup solver was used, False otherwise.

    Attributes
    ----------
    solved_globally
    results_list
    scaled_violations
    violating_param_realizations
    auxiliary_param_values
    variable_values
    found_violation
    time_out
    subsolver_error
    discrete_set_scenario_index
    backup_solver_used
    """

    def __init__(
        self,
        solved_globally,
        results_list=None,
        scaled_violations=None,
        violating_param_realization=None,
        auxiliary_param_values=None,
        variable_values=None,
        found_violation=None,
        time_out=None,
        subsolver_error=None,
        discrete_set_scenario_index=None,
        backup_solver_used=None,
    ):
        """Initialize self (see class docstring)."""
        self.results_list = results_list
        self.solved_globally = solved_globally
        self.scaled_violations = scaled_violations
        self.violating_param_realization = violating_param_realization
        self.auxiliary_param_values = auxiliary_param_values
        self.variable_values = variable_values
        self.found_violation = found_violation
        self.time_out = time_out
        self.subsolver_error = subsolver_error
        self.discrete_set_scenario_index = discrete_set_scenario_index
        self.backup_solver_used = backup_solver_used

    def termination_acceptable(self, acceptable_terminations):
        """
        Return True if termination condition for at least
        one result in `self.results_list` is in list
        of pre-specified acceptable terminations, False otherwise.

        Parameters
        ----------
        acceptable_terminations : set of pyomo.opt.TerminationCondition
            Acceptable termination conditions.

        Returns
        -------
        bool
        """
        return any(
            res.solver.termination_condition in acceptable_terminations
            for res in self.results_list
        )


class DiscreteSeparationSolveCallResults:
    """
    Container for results of solve attempt for single separation
    problem.

    Parameters
    ----------
    solved_globally : bool
        True if separation problems solved to global optimality,
        False otherwise.
    solver_call_results : dict
        Mapping from discrete uncertainty set scenario list
        indexes to solver call results for separation problems
        subject to the scenarios.
    second_stage_ineq_con : Constraint
        Separation problem second-stage inequality constraint for which
        `self` was generated.
    backup_solver_used : bool
        True if backup solver was used to solve the problem,
        False otherwise.

    Attributes
    ----------
    solved_globally
    solver_call_results
    second_stage_ineq_con
    backup_solver_used
    """

    def __init__(
        self,
        solved_globally,
        solver_call_results=None,
        second_stage_ineq_con=None,
        backup_solver_used=None,
    ):
        """Initialize self (see class docstring)."""
        self.solved_globally = solved_globally
        self.solver_call_results = solver_call_results
        self.second_stage_ineq_con = second_stage_ineq_con
        self.backup_solver_used = backup_solver_used

    @property
    def time_out(self):
        """
        bool : True if there is a time out status for at least one of
        the ``SeparationSolveCallResults`` objects listed in `self`,
        False otherwise.
        """
        return any(res.time_out for res in self.solver_call_results.values())

    @property
    def subsolver_error(self):
        """
        bool : True if there is a subsolver error status for all
        of the ``SeparationSolveCallResults`` objects listed
        in `self`, False otherwise.
        """
        return all(res.subsolver_error for res in self.solver_call_results.values())


class SeparationLoopResults:
    """
    Container for results of all separation problems solved
    to a single desired optimality target (local or global).

    Parameters
    ----------
    solved_globally : bool
        True if separation problems were solved to global optimality,
        False otherwise.
    solver_call_results : ComponentMap
        Mapping from second-stage inequality constraints to corresponding
        ``SeparationSolveCallResults`` objects.
    worst_case_ss_ineq_con : None or Constraint
        Second-stage inequality constraint mapped to
        ``SeparationSolveCallResults``
        object in `self` corresponding to maximally violating
        separation problem solution.
    all_discrete_scenarios_exhausted : bool, optional
        For problems with discrete uncertainty sets,
        True if all scenarios were explicitly accounted for in master
        (which occurs if there have been
        as many PyROS iterations as there are scenarios in the set)
        False otherwise.

    Attributes
    ----------
    solved_globally : bool
        True if global solver was used, False otherwise.
    solver_call_results : ComponentMap
        Mapping from second-stage inequality constraints to corresponding
        ``SeparationSolveCallResults`` objects.
    worst_case_ss_ineq_con : None or ConstraintData
        Worst-case second-stage inequality constraint.
    all_discrete_scenarios_exhausted : bool
        True if all scenarios of the discrete set were exhausted
        already explicitly accounted for in the master problems,
        False otherwise.
    """

    def __init__(
        self,
        solved_globally,
        solver_call_results,
        worst_case_ss_ineq_con,
        all_discrete_scenarios_exhausted=False,
    ):
        """Initialize self (see class docstring)."""
        self.solver_call_results = solver_call_results
        self.solved_globally = solved_globally
        self.worst_case_ss_ineq_con = worst_case_ss_ineq_con
        self.all_discrete_scenarios_exhausted = all_discrete_scenarios_exhausted

    @property
    def backup_solver_used(self):
        """
        bool : True if a backup solver was used to obtain any of
        the separation results contained in ``self``,
        False otherwise.
        """
        return any(
            solver_call_res.backup_solver_used
            for solver_call_res in self.solver_call_results.values()
        )

    @property
    def found_violation(self):
        """
        bool : True if separation solution for at least one
        ``SeparationSolveCallResults`` object listed in self
        was reported to violate its corresponding second-stage
        inequality constraint, False otherwise.
        """
        return any(
            solver_call_res.found_violation
            for solver_call_res in self.solver_call_results.values()
        )

    @property
    def violating_param_realization(self):
        """
        None or list of float : Uncertain parameter values for
        for maximally violating separation problem solution,
        specified according to solver call results object
        listed in self at index ``self.worst_case_ss_ineq_con``.
        If ``self.worst_case_ss_ineq_con`` is not specified,
        then None is returned.
        """
        if self.worst_case_ss_ineq_con is not None:
            return self.solver_call_results[
                self.worst_case_ss_ineq_con
            ].violating_param_realization
        else:
            return None

    @property
    def auxiliary_param_values(self):
        """
        None or list of float : Auxiliary parameter values for the
        maximially violating separation problem solution.
        """
        if self.worst_case_ss_ineq_con is not None:
            return self.solver_call_results[
                self.worst_case_ss_ineq_con
            ].auxiliary_param_values
        else:
            return None

    @property
    def scaled_violations(self):
        """
        None or ComponentMap : Scaled second-stage inequality
        constraint violations
        for maximally violating separation problem solution,
        specified according to solver call results object
        listed in self at index ``self.worst_case_ss_ineq_con``.
        If ``self.worst_case_ss_ineq_con`` is not specified,
        then None is returned.
        """
        if self.worst_case_ss_ineq_con is not None:
            return self.solver_call_results[
                self.worst_case_ss_ineq_con
            ].scaled_violations
        else:
            return None

    @property
    def violating_separation_variable_values(self):
        """
        None or ComponentMap : Second-stage and state variable values
        for maximally violating separation problem solution,
        specified according to solver call results object
        listed in self at index ``self.worst_case_ss_ineq_con``.
        If ``self.worst_case_ss_ineq_con`` is not specified,
        then None is returned.
        """
        if self.worst_case_ss_ineq_con is not None:
            return self.solver_call_results[self.worst_case_ss_ineq_con].variable_values
        else:
            return None

    @property
    def violated_second_stage_ineq_cons(self):
        """
        list of Constraint : Second-stage inequality constraints
        for which violation found.
        """
        return [
            con
            for con, solver_call_results in self.solver_call_results.items()
            if solver_call_results.found_violation
        ]

    @property
    def subsolver_error(self):
        """
        bool : Return True if subsolver error reported for
        at least one ``SeparationSolveCallResults`` stored in
        `self` and no violations are found, False otherwise.
        """
        return (
            any(
                solver_call_res.subsolver_error
                for solver_call_res in self.solver_call_results.values()
            )
            and not self.found_violation
        )

    @property
    def time_out(self):
        """
        bool : Return True if time out reported for
        at least one ``SeparationSolveCallResults`` stored in
        `self`, False otherwise.
        """
        return any(
            solver_call_res.time_out
            for solver_call_res in self.solver_call_results.values()
        )


class SeparationResults:
    """
    Container for results of PyROS separation problem routine.

    Parameters
    ----------
    local_separation_loop_results : None or SeparationLoopResults
        Local separation problem loop results.
    global_separation_loop_results : None or SeparationLoopResults
        Global separation problem loop results.

    Attributes
    ----------
    local_separation_loop_results : None or SeparationLoopResults
        Local separation results. If separation problems
        were not solved locally, then this attribute is set
        to None.
    global_separation_loop_results : None or SeparationLoopResults
        Global separation results. If separation problems
        were not solved globally, then this attribute is set
        to None.
    """

    def __init__(self, local_separation_loop_results, global_separation_loop_results):
        """Initialize self (see class docstring)."""
        self.local_separation_loop_results = local_separation_loop_results
        self.global_separation_loop_results = global_separation_loop_results

    @property
    def backup_local_solver_used(self):
        """
        bool : True if a backup solver was used for local separation,
        False otherwise.
        """
        return (
            self.local_separation_loop_results
            and self.local_separation_loop_results.backup_solver_used
        )

    @property
    def backup_global_solver_used(self):
        """
        bool : True if a backup solver was used for global separation,
        False otherwise.
        """
        return (
            self.global_separation_loop_results
            and self.global_separation_loop_results.backup_solver_used
        )

    @property
    def time_out(self):
        """
        bool : True if time out found for local or global
        separation loop, False otherwise.
        """
        local_time_out = (
            self.solved_locally and self.local_separation_loop_results.time_out
        )
        global_time_out = (
            self.solved_globally and self.global_separation_loop_results.time_out
        )
        return local_time_out or global_time_out

    @property
    def subsolver_error(self):
        """
        bool : True if subsolver error found for local or global
        separation loop, False otherwise.
        """
        local_subsolver_error = (
            self.solved_locally and self.local_separation_loop_results.subsolver_error
        )
        global_subsolver_error = (
            self.solved_globally and self.global_separation_loop_results.subsolver_error
        )
        return local_subsolver_error or global_subsolver_error

    @property
    def solved_locally(self):
        """
        bool : true if local separation loop was invoked,
        False otherwise.
        """
        return self.local_separation_loop_results is not None

    @property
    def solved_globally(self):
        """
        bool : True if global separation loop was invoked,
        False otherwise.
        """
        return self.global_separation_loop_results is not None

    def get_violating_attr(self, attr_name):
        """
        If separation problems solved globally, returns
        value of attribute of global separation loop results.

        Otherwise, if separation problems solved locally,
        returns value of attribute of local separation loop results.
        If local separation loop results specified, return
        value of attribute of local separation loop results.

        Otherwise, if global separation loop results specified,
        return value of attribute of global separation loop
        results.

        Otherwise, return None.

        Parameters
        ----------
        attr_name : str
            Name of attribute to be retrieved. Should be
            valid attribute name for object of type
            ``SeparationLoopResults``.

        Returns
        -------
        object
            Attribute value.
        """
        return getattr(self.main_loop_results, attr_name, None)

    @property
    def all_discrete_scenarios_exhausted(self):
        """
        bool : For problems where the uncertainty set is of type
        DiscreteScenarioSet,
        True if last master problem solved explicitly
        accounts for all scenarios in the uncertainty set,
        False otherwise.
        """
        return self.get_violating_attr("all_discrete_scenarios_exhausted")

    @property
    def worst_case_ss_ineq_con(self):
        """
        ConstraintData : Second-stage inequality constraint
        corresponding to the
        separation solution chosen for the next master problem.
        """
        return self.get_violating_attr("worst_case_ss_ineq_con")

    @property
    def main_loop_results(self):
        """
        SeparationLoopResults : Main separation loop results.
        In particular, this is considered to be the global
        loop result if solved globally, and the local loop
        results otherwise.
        """
        if self.solved_globally:
            return self.global_separation_loop_results
        return self.local_separation_loop_results

    @property
    def found_violation(self):
        """
        bool : True if ``found_violation`` attribute for
        main separation loop results is True, False otherwise.
        """
        found_viol = self.get_violating_attr("found_violation")
        if found_viol is None:
            found_viol = False
        return found_viol

    @property
    def violating_param_realization(self):
        """
        None or list of float : Uncertain parameter values
        for maximally violating separation problem solution
        reported in local or global separation loop results.
        If no such solution found, (i.e. ``worst_case_ss_ineq_con``
        set to None for both local and global loop results),
        then None is returned.
        """
        return self.get_violating_attr("violating_param_realization")

    @property
    def auxiliary_param_values(self):
        """
        None or list of float: Auxiliary parameter values accompanying
        `self.violating_param_realization`.
        """
        return self.get_violating_attr("auxiliary_param_values")

    @property
    def scaled_violations(self):
        """
        None or ComponentMap :
        Scaled second-stage inequality constraint violations
        for maximally violating separation problem solution
        reported in local or global separation loop results.
        If no such solution found, (i.e. ``worst_case_ss_ineq_con``
        set to None for both local and global loop results),
        then None is returned.
        """
        return self.get_violating_attr("scaled_violations")

    @property
    def violating_separation_variable_values(self):
        """
        None or ComponentMap : Second-stage and state variable values
        for maximally violating separation problem solution
        reported in local or global separation loop results.
        If no such solution found, (i.e. ``worst_case_ss_ineq_con``
        set to None for both local and global loop results),
        then None is returned.
        """
        return self.get_violating_attr("violating_separation_variable_values")

    @property
    def violated_second_stage_ineq_cons(self):
        """
        Return list of violated second-stage inequality constraints.
        """
        return self.get_violating_attr("violated_second_stage_ineq_cons")

    @property
    def robustness_certified(self):
        """
        bool : Return True if separation results certify that
        first-stage solution is robust, False otherwise.
        """
        assert self.solved_locally or self.solved_globally

        if self.time_out or self.subsolver_error:
            return False

        if self.solved_locally:
            heuristically_robust = (
                not self.local_separation_loop_results.found_violation
            )
        else:
            heuristically_robust = None

        if self.solved_globally:
            is_robust = not self.global_separation_loop_results.found_violation
        else:
            # global separation bypassed, either
            # because uncertainty set is discrete
            # or user opted to bypass global separation
            is_robust = heuristically_robust

        return is_robust



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.contrib.piecewise.tests import models
import pyomo.contrib.piecewise.tests.common_tests as ct
import pyomo.contrib.piecewise.tests.common_inner_repn_tests as inner_repn_tests
from pyomo.core.base import TransformationFactory
from pyomo.environ import SolverFactory, Var, Constraint
from pyomo.gdp import Disjunction, Disjunct
from pyomo.core.expr.compare import assertExpressionsEqual


# Test the nested inner repn gdp model using the common_tests code
class TestTransformPiecewiseModelToNestedInnerRepnGDP(unittest.TestCase):
    # Check the structure of the log PWLF Block
    def check_pw_log(self, m):
        z = m.pw_log.get_transformation_var(m.log_expr)
        self.assertIsInstance(z, Var)
        # Now we can use those Vars to check on what the transformation created
        log_block = z.parent_block()

        # Not using ct.check_trans_block_structure() because these are slightly
        # different
        # Two top-level disjuncts
        self.assertEqual(len(log_block.component_map(Disjunct)), 2)
        # One disjunction
        self.assertEqual(len(log_block.component_map(Disjunction)), 1)
        # The 'z' var (that we will substitute in for the function being
        # approximated) is here:
        self.assertEqual(len(log_block.component_map(Var)), 1)
        self.assertIsInstance(log_block.substitute_var, Var)

        # Check the tree structure, which should be heavier on the right
        # Parent disjunction
        self.assertIsInstance(log_block.disj, Disjunction)
        self.assertEqual(len(log_block.disj.disjuncts), 2)

        # Left disjunct with constraints
        self.assertIsInstance(log_block.d_l, Disjunct)
        inner_repn_tests.check_log_disjunct(
            self, log_block.d_l, (1, 3), m.f1, log_block.substitute_var, m.x
        )

        # Right disjunct with disjunction
        self.assertIsInstance(log_block.d_r, Disjunct)
        self.assertIsInstance(log_block.d_r.inner_disjunction_r, Disjunction)
        self.assertEqual(len(log_block.d_r.inner_disjunction_r.disjuncts), 2)

        # Left and right child disjuncts with constraints
        self.assertIsInstance(log_block.d_r.d_l, Disjunct)
        inner_repn_tests.check_log_disjunct(
            self, log_block.d_r.d_l, (3, 6), m.f2, log_block.substitute_var, m.x
        )
        self.assertIsInstance(log_block.d_r.d_r, Disjunct)
        inner_repn_tests.check_log_disjunct(
            self, log_block.d_r.d_r, (6, 10), m.f3, log_block.substitute_var, m.x
        )

        # Check that this also became the objective
        self.assertIs(m.obj.expr.expr, log_block.substitute_var)

    # Check the structure of the paraboloid PWLF block
    def check_pw_paraboloid(self, m):
        z = m.pw_paraboloid.get_transformation_var(m.paraboloid_expr)
        self.assertIsInstance(z, Var)
        paraboloid_block = z.parent_block()

        # Two top-level disjuncts
        self.assertEqual(len(paraboloid_block.component_map(Disjunct)), 2)
        # One disjunction
        self.assertEqual(len(paraboloid_block.component_map(Disjunction)), 1)
        # The 'z' var (that we will substitute in for the function being
        # approximated) is here:
        self.assertEqual(len(paraboloid_block.component_map(Var)), 1)
        self.assertIsInstance(paraboloid_block.substitute_var, Var)

        # This one should have an even tree with four leaf disjuncts
        disjuncts_dict = {
            paraboloid_block.d_l.d_l: ([(0, 1), (0, 4), (3, 4)], m.g1),
            paraboloid_block.d_l.d_r: ([(0, 1), (3, 4), (3, 1)], m.g1),
            paraboloid_block.d_r.d_l: ([(3, 4), (3, 7), (0, 7)], m.g2),
            paraboloid_block.d_r.d_r: ([(0, 7), (0, 4), (3, 4)], m.g2),
        }
        for d, (pts, f) in disjuncts_dict.items():
            inner_repn_tests.check_paraboloid_disjunct(
                self, d, pts, f, paraboloid_block.substitute_var, m.x1, m.x2
            )

        # And check the substitute Var is in the objective now.
        self.assertIs(m.indexed_c[0].body.args[0].expr, paraboloid_block.substitute_var)

    # Test methods using the common_tests.py code. Copied in from test_inner_repn_gdp.py.
    def test_transformation_do_not_descend(self):
        ct.check_transformation_do_not_descend(
            self, 'contrib.piecewise.nested_inner_repn_gdp'
        )

    def test_transformation_PiecewiseLinearFunction_targets(self):
        ct.check_transformation_PiecewiseLinearFunction_targets(
            self, 'contrib.piecewise.nested_inner_repn_gdp'
        )

    def test_descend_into_expressions(self):
        ct.check_descend_into_expressions(
            self, 'contrib.piecewise.nested_inner_repn_gdp'
        )

    def test_descend_into_expressions_constraint_target(self):
        ct.check_descend_into_expressions_constraint_target(
            self, 'contrib.piecewise.nested_inner_repn_gdp'
        )

    def test_descend_into_expressions_objective_target(self):
        ct.check_descend_into_expressions_objective_target(
            self, 'contrib.piecewise.nested_inner_repn_gdp'
        )

    # Check the solution of the log(x) model
    @unittest.skipUnless(SolverFactory('gurobi').available(), 'Gurobi is not available')
    @unittest.skipUnless(SolverFactory('gurobi').license_is_valid(), 'No license')
    def test_solve_log_model(self):
        m = models.make_log_x_model()
        TransformationFactory("contrib.piecewise.nested_inner_repn_gdp").apply_to(m)
        TransformationFactory("gdp.bigm").apply_to(m)
        SolverFactory("gurobi").solve(m)
        ct.check_log_x_model_soln(self, m)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.contrib.pyros.pyros import PyROS
from pyomo.contrib.pyros.util import ObjectiveType, pyrosTerminationCondition
from pyomo.contrib.pyros.uncertainty_sets import (
    UncertaintySet,
    EllipsoidalSet,
    PolyhedralSet,
    CardinalitySet,
    BudgetSet,
    DiscreteScenarioSet,
    FactorModelSet,
    BoxSet,
    IntersectionSet,
    AxisAlignedEllipsoidalSet,
)



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import subprocess
import sys
from os.path import join, exists, splitext

import pyomo.common.unittest as unittest

from pyomo.common.fileutils import this_file_dir
from pyomo.common.tempfiles import TempfileManager

import pyomo.environ
from pyomo.opt import SolverFactory
from pyomo.core import ConcreteModel, Var, Objective, Constraint

import pyomo.solvers.plugins.solvers.SCIPAMPL

currdir = this_file_dir()
deleteFiles = True


@unittest.skipIf(
    'pypy_version_info' in dir(sys), "Skip SCIPAMPL tests on Pypy due to performance"
)
class Test(unittest.TestCase):
    def setUp(self):
        scip = SolverFactory('scip', solver_io='nl')
        type(scip)._known_versions = {}
        TempfileManager.push()

        self.patch_run = unittest.mock.patch(
            'pyomo.solvers.plugins.solvers.SCIPAMPL.subprocess.run'
        )
        # Executable cannot be partially mocked since it creates a PathData object.
        self.patch_path = unittest.mock.patch.object(
            pyomo.common.fileutils.PathData, 'path', autospec=True
        )
        self.patch_available = unittest.mock.patch.object(
            pyomo.common.fileutils.PathData, 'available', autospec=True
        )

        self.run = self.patch_run.start()
        self.path = self.patch_path.start()
        self.available = self.patch_available.start()

        self.executable_paths = {
            "scip": join(currdir, "scip"),
            "scipampl": join(currdir, "scipampl"),
        }

    def tearDown(self):
        self.patch_run.stop()
        self.patch_path.stop()
        self.patch_available.stop()

        TempfileManager.pop(remove=deleteFiles or self.currentTestPassed())

    def generate_stdout(self, solver, version):
        if solver == "scip":
            # Template from SCIP 8.0.0
            stdout = (
                "SCIP version {} [precision: 8 byte] [memory: block] [mode: optimized] [LP solver: SoPlex 6.0.0] [GitHash: d9b84b0709]\n"
                "Copyright (C) 2002-2021 Konrad-Zuse-Zentrum fuer Informationstechnik Berlin (ZIB)\n"
                "\n"
                "External libraries:\n"
                "   SoPlex 6.0.0    Linear Programming Solver developed at Zuse Institute Berlin (soplex.zib.de) [GitHash: f5cfa86b]"
            )

            # Template from SCIPAMPL 7.0.3
        elif solver == "scipampl":
            stdout = (
                "SCIP version {} [precision: 8 byte] [memory: block] [mode: optimized] [LP solver: SoPlex 5.0.2] [GitHash: 74c11e60cd]\n"
                "Copyright (C) 2002-2021 Konrad-Zuse-Zentrum fuer Informationstechnik Berlin (ZIB)\n"
                "\n"
                "External libraries:\n"
                " Readline 8.0         GNU library for command line editing (gnu.org/s/readline)"
            )
        else:
            raise ValueError("Unsupported solver for stdout generation.")

        version = ".".join(str(e) for e in version[:3])
        return stdout.format(version)

    def set_solvers(self, scip=(8, 0, 0, 0), scipampl=(7, 0, 3, 0), fail=True):
        executables = {"scip": scip, "scipampl": scipampl}

        def get_executable(*args, **kwargs):
            name = args[0]._registered_name
            if name in executables:
                if executables[name]:
                    return self.executable_paths[name]
                else:
                    return None
            elif fail:
                self.fail("Solver creation looked up a non scip executable.")
            else:
                return False

        def get_available(*args, **kwargs):
            name = args[0]._registered_name
            if name in executables:
                return executables[name] is not None
            elif fail:
                self.fail("Solver creation looked up a non scip executable.")
            else:
                return False

        def run(args, **kwargs):
            for solver_name, solver_version in executables.items():
                if not args[0].endswith(solver_name):
                    continue
                if solver_version is None:
                    raise FileNotFoundError()
                else:
                    return subprocess.CompletedProcess(
                        args, 0, self.generate_stdout(solver_name, solver_version), None
                    )
            if fail:
                self.fail("Solver creation looked up a non scip executable.")

        self.path.side_effect = get_executable
        self.available.side_effect = get_available
        self.run.side_effect = run

    def test_scip_available(self):
        self.set_solvers()
        scip = SolverFactory('scip', solver_io='nl')
        scip_executable = scip.executable()
        self.assertIs(scip_executable, self.executable_paths["scip"])
        # only one call to path expected, since no check for SCIPAMPL is needed
        self.assertEqual(1, self.path.call_count)
        self.assertEqual(1, self.run.call_count)
        self.available.assert_called()

        # version should now be cached
        scip.executable()
        self.assertEqual(1, self.run.call_count)

        self.assertTrue(scip.available())

    def test_scipampl_fallback(self):
        self.set_solvers(scip=(7, 0, 3, 0))
        scip = SolverFactory('scip', solver_io='nl')
        scip_executable = scip.executable()
        self.assertIs(scip_executable, self.executable_paths["scipampl"])

        # get SCIP and SCIPAMPL paths
        self.assertEqual(2, self.path.call_count)
        # only check SCIP version
        self.assertEqual(1, self.run.call_count)
        self.available.assert_called()

        self.assertEqual((7, 0, 3, 0), scip._get_version())
        # also check SCIPAMPL version
        self.assertEqual(2, self.run.call_count)

        # versions should now be cached
        scip.executable()
        self.assertEqual(2, self.run.call_count)

        self.assertTrue(scip.available())

    def test_no_scip(self):
        self.set_solvers(scip=None)
        scip = SolverFactory('scip', solver_io='nl')
        scip_executable = scip.executable()
        self.assertIs(scip_executable, self.executable_paths["scipampl"])

        # get scipampl path
        self.assertEqual(1, self.path.call_count)
        # cannot check SCIP version
        self.assertEqual(0, self.run.call_count)
        self.available.assert_called()

        self.assertEqual((7, 0, 3, 0), scip._get_version())
        # also check SCIPAMPL version
        self.assertEqual(1, self.run.call_count)

        # versions should now be cached
        scip.executable()
        self.assertEqual(1, self.run.call_count)

        self.assertTrue(scip.available())

    def test_no_fallback(self):
        self.set_solvers(scip=None, scipampl=None)
        scip = SolverFactory('scip', solver_io='nl')
        self.assertFalse(scip.available())
        self.assertIsNone(scip.executable())

        # cannot check SCIP versions
        self.assertEqual(0, self.run.call_count)
        self.available.assert_called()

    def test_scip_solver_options(self):
        self.set_solvers(fail=False)
        scip = SolverFactory('scip', solver_io='nl')
        m = self.model = ConcreteModel()
        m.v = Var()
        m.o = Objective(expr=m.v)
        m.c = Constraint(expr=m.v >= 1)

        # cache version and reset mock
        scip._get_version()
        self.run.reset_mock()

        # SCIP is not actually called
        with self.assertRaises(FileNotFoundError) as cm:
            scip.solve(m, timelimit=10)

        # SCIP calls should have 3 options
        args = self.run.call_args[0][0]
        self.assertEqual(3, len(args))
        # check scip call
        self.assertEqual(self.executable_paths["scip"], args[0])
        # check for nl file existence
        self.assertTrue(exists(args[1] + ".nl"))
        # check proper sol filename
        self.assertEqual(args[1] + ".sol", cm.exception.filename)
        # check -AMPL option
        self.assertEqual("-AMPL", args[2])
        # check options file
        options_dir = self.run.call_args[1]['cwd']
        self.assertTrue(exists(options_dir + "/scip.set"))
        with open(options_dir + "/scip.set", 'r') as options:
            self.assertEqual(["limits/time = 10\n"], options.readlines())

    def test_scipampl_solver_options(self):
        self.set_solvers(scip=None, fail=False)
        scip = SolverFactory('scip', solver_io='nl')
        m = self.model = ConcreteModel()
        m.v = Var()
        m.o = Objective(expr=m.v)
        m.c = Constraint(expr=m.v >= 1)

        # cache version and reset mock
        scip._get_version()
        self.run.reset_mock()

        # SCIP is not actually called
        with self.assertRaises(FileNotFoundError) as cm:
            scip.solve(m, timelimit=10, options={'numerics/feastol': 1e-09})

        # check scip call
        args = self.run.call_args[0][0]
        self.assertEqual(self.executable_paths["scipampl"], args[0])
        # check for nl file existence
        self.assertTrue(exists(args[1]))
        root, ext = splitext(args[1])
        self.assertEqual(".nl", ext)
        # check proper sol filename
        self.assertEqual(root + ".sol", cm.exception.filename)
        # check -AMPL option
        self.assertEqual("-AMPL", args[2])
        # check options file
        options_dir = self.run.call_args[1].get('cwd', None)

        if options_dir is not None and exists(options_dir + "/scip.set"):
            # SCIPAMPL call should have 3 options
            self.assertEqual(3, len(args))
            options_file = options_dir + "/scip.set"
        else:
            # SCIPAMPL call should have 4 options
            self.assertEqual(4, len(args))
            # SCIPAMPL can also receive the options file as the fourth command line argument
            options_file = args[3]
            self.assertTrue(exists(options_file))

        with open(options_file, 'r') as options:
            lines = options.readlines()
            self.assertIn("numerics/feastol = 1e-09\n", lines)
            self.assertIn("limits/time = 10\n", lines)


if __name__ == "__main__":
    deleteFiles = False
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from pyomo.common.collections import ComponentMap
from pyomo.common.errors import MouseTrap
from pyomo.core.expr.expr_common import ExpressionType
from pyomo.core.expr.visitor import StreamBasedExpressionVisitor
import pyomo.core.expr as EXPR
from pyomo.core.base import (
    Binary,
    Constraint,
    ConstraintList,
    NonNegativeIntegers,
    VarList,
    value,
)
import pyomo.core.base.boolean_var as BV
from pyomo.core.base.expression import ScalarExpression, ExpressionData
from pyomo.core.base.param import ScalarParam, ParamData
from pyomo.core.base.var import ScalarVar, VarData
from pyomo.gdp.disjunct import AutoLinkedBooleanVar, Disjunct, Disjunction


def _dispatch_boolean_const(visitor, node):
    return False, 1 if node.value else 0


def _dispatch_boolean_var(visitor, node):
    if node not in visitor.boolean_to_binary_map:
        binary = node.get_associated_binary()
        if binary is not None:
            visitor.boolean_to_binary_map[node] = binary
        else:
            z = visitor.z_vars.add()
            visitor.boolean_to_binary_map[node] = z
            node.associate_binary_var(z)
    if node.fixed:
        visitor.boolean_to_binary_map[node].fixed = True
        visitor.boolean_to_binary_map[node].set_value(
            int(node.value) if node.value is not None else None, skip_validation=True
        )
    return False, visitor.boolean_to_binary_map[node]


def _dispatch_var(visitor, node):
    return False, node


def _dispatch_param(visitor, node):
    return False, node


def _dispatch_expression(visitor, node):
    return False, node.expr


def _before_relational_expr(visitor, node):
    raise MouseTrap(
        "The RelationalExpression '%s' was used as a Boolean term "
        "in a logical proposition. This is not yet supported "
        "when transforming to disjunctive form." % node
    )


def _dispatch_not(visitor, node, a):
    # z == !a
    if a not in visitor.expansions:
        z = visitor.z_vars.add()
        visitor.constraints.add(z == 1 - a)
        visitor.expansions[a] = z
    return visitor.expansions[a]


def _dispatch_implication(visitor, node, a, b):
    # z == !a v b
    return _dispatch_or(visitor, node, 1 - a, b)


def _dispatch_equivalence(visitor, node, a, b):
    # z == (!a v b) ^ (a v !b)
    return _dispatch_and(
        visitor,
        node,
        _dispatch_or(visitor, node, 1 - a, b),
        _dispatch_or(visitor, node, a, 1 - b),
    )


def _dispatch_and(visitor, node, *args):
    # z == a ^ b ^ ...
    z = visitor.z_vars.add()
    for arg in args:
        visitor.constraints.add(arg >= z)
    visitor.constraints.add(len(args) - sum(args) >= 1 - z)
    return z


def _dispatch_or(visitor, node, *args):
    # z == a v b v ...
    # (!z v a v b v ...) ^ (z v !a) ^ (z v !b) ^ ...
    z = visitor.z_vars.add()
    visitor.constraints.add((1 - z) + sum(args) >= 1)
    for arg in args:
        visitor.constraints.add(z + (1 - arg) >= 1)
    return z


def _dispatch_xor(visitor, node, a, b):
    # z == a XOR b
    # This is a special case of exactly
    return _dispatch_exactly(visitor, node, 1, a, b)


def _get_integer_value(n, node):
    if n.__class__ in EXPR.native_numeric_types and int(n) == n:
        return n
    if n.__class__ not in EXPR.native_types:
        if n.is_potentially_variable():
            # [ESJ 11/22]: This is probably worth supporting sometime, but right
            # now we are abiding by what docplex allows in their 'count'
            # function. Part of supporting this will be making sure we catch
            # strict inequalities in the GDP transformations. Because if we
            # don't know that n is integer-valued we will be forced to write
            # strict inequalities instead of incrememting or decrementing by 1
            # in the disjunctions.
            raise MouseTrap(
                "The first argument '%s' to '%s' is potentially variable. "
                "This may be a mathematically coherent expression; However "
                "it is not yet supported to convert it to a disjunctive "
                "program." % (n, node)
            )
        else:
            return n
    raise ValueError(
        "The first argument to '%s' must be an integer.\n\tRecieved: %s" % (node, n)
    )


def _dispatch_exactly(visitor, node, *args):
    # z = sum(args[1:]) == args[0]
    # This is currently implemented as:
    # [sum(args[1:] = n] v [[sum(args[1:]) < n] v [sum(args[1:]) > n]]
    M = len(args) - 1
    n = _get_integer_value(args[0], node)
    sum_expr = sum(args[1:])
    equality_disj = visitor.disjuncts[len(visitor.disjuncts)]
    equality_disj.constraint = Constraint(expr=sum_expr == n)
    inequality_disj = visitor.disjuncts[len(visitor.disjuncts)]
    inequality_disj.disjunction = Disjunction(
        expr=[[sum_expr <= n - 1], [sum_expr >= n + 1]]
    )
    visitor.disjunctions[len(visitor.disjunctions)] = [equality_disj, inequality_disj]
    return equality_disj.indicator_var.get_associated_binary()


def _dispatch_atleast(visitor, node, *args):
    # z = sum[args[1:] >= n
    # This is implemented as:
    # [sum(args[1:] >= n] v [sum(args[1:] < n]
    n = _get_integer_value(args[0], node)
    sum_expr = sum(args[1:])
    atleast_disj = visitor.disjuncts[len(visitor.disjuncts)]
    less_disj = visitor.disjuncts[len(visitor.disjuncts)]
    atleast_disj.constraint = Constraint(expr=sum_expr >= n)
    less_disj.constraint = Constraint(expr=sum_expr <= n - 1)
    visitor.disjunctions[len(visitor.disjunctions)] = [atleast_disj, less_disj]
    return atleast_disj.indicator_var.get_associated_binary()


def _dispatch_atmost(visitor, node, *args):
    # z = sum[args[1:] <= n
    # This is implemented as:
    # [sum(args[1:] <= n] v [sum(args[1:] > n]
    n = _get_integer_value(args[0], node)
    sum_expr = sum(args[1:])
    atmost_disj = visitor.disjuncts[len(visitor.disjuncts)]
    more_disj = visitor.disjuncts[len(visitor.disjuncts)]
    atmost_disj.constraint = Constraint(expr=sum_expr <= n)
    more_disj.constraint = Constraint(expr=sum_expr >= n + 1)
    visitor.disjunctions[len(visitor.disjunctions)] = [atmost_disj, more_disj]
    return atmost_disj.indicator_var.get_associated_binary()


_operator_dispatcher = {}
_operator_dispatcher[EXPR.ImplicationExpression] = _dispatch_implication
_operator_dispatcher[EXPR.EquivalenceExpression] = _dispatch_equivalence
_operator_dispatcher[EXPR.NotExpression] = _dispatch_not
_operator_dispatcher[EXPR.AndExpression] = _dispatch_and
_operator_dispatcher[EXPR.OrExpression] = _dispatch_or
_operator_dispatcher[EXPR.XorExpression] = _dispatch_xor
_operator_dispatcher[EXPR.ExactlyExpression] = _dispatch_exactly
_operator_dispatcher[EXPR.AtLeastExpression] = _dispatch_atleast
_operator_dispatcher[EXPR.AtMostExpression] = _dispatch_atmost

_before_child_dispatcher = {}
_before_child_dispatcher[EXPR.BooleanConstant] = _dispatch_boolean_const
_before_child_dispatcher[BV.ScalarBooleanVar] = _dispatch_boolean_var
_before_child_dispatcher[BV.BooleanVarData] = _dispatch_boolean_var
_before_child_dispatcher[AutoLinkedBooleanVar] = _dispatch_boolean_var
_before_child_dispatcher[ParamData] = _dispatch_param
_before_child_dispatcher[ScalarParam] = _dispatch_param
# for the moment, these are all just so we can get good error messages when we
# don't handle them:
_before_child_dispatcher[ScalarVar] = _dispatch_var
_before_child_dispatcher[VarData] = _dispatch_var
_before_child_dispatcher[ExpressionData] = _dispatch_expression
_before_child_dispatcher[ScalarExpression] = _dispatch_expression


class LogicalToDisjunctiveVisitor(StreamBasedExpressionVisitor):
    """Converts BooleanExpressions to Linear (MIP) representation

    This converter eschews conjunctive normal form, and instead follows
    the well-trodden MINLP path of factorable programming.

    """

    def __init__(self):
        super().__init__()
        self.z_vars = VarList(domain=Binary)
        self.z_vars.construct()
        self.constraints = ConstraintList()
        self.disjuncts = Disjunct(NonNegativeIntegers, concrete=True)
        self.disjunctions = Disjunction(NonNegativeIntegers)
        self.disjunctions.construct()
        self.expansions = ComponentMap()
        self.boolean_to_binary_map = ComponentMap()

    def initializeWalker(self, expr):
        walk, result = self.beforeChild(None, expr, 0)
        if not walk:
            return False, self.finalizeResult(result)
        return True, expr

    def beforeChild(self, node, child, child_idx):
        if child.__class__ in EXPR.native_types:
            if child.__class__ is bool:
                # If we encounter a bool, we are going to need to treat it as
                # binary explicitly because we are finally pedantic enough in the
                # expression system to not allow some of the mixing we will need
                # (like summing a LinearExpression with a bool)
                return False, int(child)
            return False, child

        if child.is_numeric_type():
            # Just pass it through, we'll figure it out later
            return False, child
        if child.is_expression_type(ExpressionType.RELATIONAL):
            # Eventually we'll handle these. Right now we set a MouseTrap
            return _before_relational_expr(self, child)

        if not child.is_expression_type() or child.is_named_expression_type():
            return _before_child_dispatcher[child.__class__](self, child)

        return True, None

    def exitNode(self, node, data):
        return _operator_dispatcher[node.__class__](self, node, *data)

    def finalizeResult(self, result):
        # This LogicalExpression must evaluate to True (but note that we cannot
        # fix this variable to 1 since this logical expression could be living
        # on a Disjunct and later need to be relaxed.)
        expr = result >= 1
        if expr.__class__ is bool:
            self.constraints.add(Constraint.Feasible if expr else Constraint.Infeasible)
        else:
            self.constraints.add(expr)
        return result



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

from collections import namedtuple
from pyomo.common.timing import HierarchicalTimer
from pyomo.common.config import ConfigBlock
from pyomo.util.subsystems import create_subsystem_block


class SquareNlpSolverBase:
    """A base class for NLP solvers that act on a square system
    of equality constraints.

    """

    # Ideally, this ConfigBlock would contain options that are valid for any
    # square NLP solver. However, no such options seem to exist while
    # preserving the names of the SciPy function arguments. E.g., tolerance
    # is "tol" in some solvers and "xtol" in others, and some solvers
    # support "maxiter" while others support "maxfev". It may be useful to
    # attempt some standardization by, e.g., mapping tol->xtol, then
    # specifying "universal" options here, but this can happen at a later
    # date as these solvers see more use.
    OPTIONS = ConfigBlock()

    def __init__(self, nlp, timer=None, options=None):
        """
        Arguments
        ---------
        nlp: ExtendedNLP
            An instance of ExtendedNLP that will be solved.
            ExtendedNLP is required to ensure that the NLP has equal
            numbers of primal variables and equality constraints.

        """
        if timer is None:
            timer = HierarchicalTimer()
        if options is None:
            options = {}
        self.options = self.OPTIONS(options)

        self._timer = timer
        self._nlp = nlp
        self._function_values = None
        self._jacobian = None

        if self._nlp.n_eq_constraints() != self._nlp.n_primals():
            raise RuntimeError(
                "Cannot construct a square solver for an NLP that"
                " does not have the same numbers of variables as"
                " equality constraints. Got %s variables and %s"
                " equalities." % (self._nlp.n_primals(), self._nlp.n_eq_constraints())
            )
        # Checking for a square system of equalities is easy, but checking
        # bounds is a little difficult. We don't know how an NLP will
        # implement bounds (no bound could be None, np.nan, or np.inf),
        # so it is annoying to check that bounds are not present.
        # Instead, we just ignore bounds, and the user must know that the
        # result of this solver is not guaranteed to respect bounds.
        # While it is easier to check that inequalities are absent,
        # for consistency, we take the same approach and simply ignore
        # them.

    def solve(self, x0=None):
        # the NLP has a natural initial guess - the cached primal
        # values. x0 may be provided if a different initial guess
        # is desired.
        raise NotImplementedError(
            "%s has not implemented the solve method" % self.__class__
        )

    def evaluate_function(self, x0):
        # NOTE: NLP object should handle any caching
        self._nlp.set_primals(x0)
        values = self._nlp.evaluate_eq_constraints()
        return values

    def evaluate_jacobian(self, x0):
        # NOTE: NLP object should handle any caching
        self._nlp.set_primals(x0)
        self._jacobian = self._nlp.evaluate_jacobian_eq(out=self._jacobian)
        return self._jacobian


class DenseSquareNlpSolver(SquareNlpSolverBase):
    """A square NLP solver that uses a dense Jacobian"""

    def evaluate_jacobian(self, x0):
        sparse_jac = super().evaluate_jacobian(x0)
        dense_jac = sparse_jac.toarray()
        return dense_jac


class ScalarDenseSquareNlpSolver(DenseSquareNlpSolver):
    # A base class for solvers for scalar equations.
    # Not intended to be instantiated directly. Instead,
    # NewtonNlpSolver or SecantNewtonNlpSolver should be used.

    def __init__(self, nlp, timer=None, options=None):
        super().__init__(nlp, timer=timer, options=options)
        if nlp.n_primals() != 1:
            raise RuntimeError(
                "Cannot use the scipy.optimize.newton solver on an NLP with"
                " more than one variable and equality constraint. Got %s"
                " primals. Please use RootNlpSolver or FsolveNlpSolver instead."
            )



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.environ as pyo
import pyomo.contrib.iis.mis as mis
from pyomo.contrib.iis.mis import _get_constraint
from pyomo.common.tempfiles import TempfileManager

import logging
import os


def _get_infeasible_model():
    m = pyo.ConcreteModel("trivial4test")
    m.x = pyo.Var(within=pyo.Binary)
    m.y = pyo.Var(within=pyo.NonNegativeReals)

    m.c1 = pyo.Constraint(expr=m.y <= 100.0 * m.x)
    m.c2 = pyo.Constraint(expr=m.y <= -100.0 * m.x)
    m.c3 = pyo.Constraint(expr=m.x >= 0.5)

    m.o = pyo.Objective(expr=-m.y)

    return m


def _get_feasible_model():
    m = pyo.ConcreteModel("Trivial Feasible Quad")
    m.x = pyo.Var([1, 2], bounds=(0, 1))
    m.y = pyo.Var(bounds=(0, 1))
    m.c = pyo.Constraint(expr=m.x[1] * m.x[2] >= -1)
    m.d = pyo.Constraint(expr=m.x[1] + m.y >= 1)

    return m


class TestMIS(unittest.TestCase):
    @unittest.skipUnless(
        pyo.SolverFactory("ipopt").available(exception_flag=False),
        "ipopt not available",
    )
    def test_write_mis_ipopt(self):
        _test_mis("ipopt")

    def test__get_constraint_errors(self):
        # A not-completely-cynical way to get the coverage up.
        m = _get_infeasible_model()  # not modified
        fct = _get_constraint

        m.foo_slack_plus_ = pyo.Var()
        self.assertRaises(RuntimeError, fct, m, m.foo_slack_plus_)
        m.foo_slack_minus_ = pyo.Var()
        self.assertRaises(RuntimeError, fct, m, m.foo_slack_minus_)
        m.foo_bar = pyo.Var()
        self.assertRaises(RuntimeError, fct, m, m.foo_bar)

    def test_feasible_model(self):
        m = _get_feasible_model()
        opt = pyo.SolverFactory("ipopt")
        self.assertRaises(Exception, mis.compute_infeasibility_explanation, m, opt)


def _check_output(file_name):
    # pretty simple check for now
    with open(file_name, "r+") as file1:
        lines = file1.readlines()
    trigger = "Constraints / bounds in MIS:"
    nugget = "lb of var y"
    live = False  # (long i)
    found_nugget = False
    for line in lines:
        if trigger in line:
            live = True
        if live:
            if nugget in line:
                found_nugget = True
    if not found_nugget:
        raise RuntimeError(f"Did not find '{nugget}' after '{trigger}' in output")
    else:
        pass


def _test_mis(solver_name):
    m = _get_infeasible_model()
    opt = pyo.SolverFactory(solver_name)

    # This test seems to fail on Windows as it unlinks the tempfile, so live with it
    #    On a Windows machine, we will not use a temp dir and just try to delete the log file
    if os.name == "nt":
        file_name = f"_test_mis_{solver_name}.log"
        logger = logging.getLogger(f"test_mis_{solver_name}")
        logger.setLevel(logging.INFO)
        fh = logging.FileHandler(file_name)
        fh.setLevel(logging.DEBUG)
        logger.addHandler(fh)

        mis.compute_infeasibility_explanation(m, opt, logger=logger)
        _check_output(file_name)
        # os.remove(file_name) cannot remove it on Windows. Still in use.

    else:  # not windows
        with TempfileManager.new_context() as tmpmgr:
            tmp_path = tmpmgr.mkdtemp()
            file_name = os.path.join(tmp_path, f"_test_mis_{solver_name}.log")
            logger = logging.getLogger(f"test_mis_{solver_name}")
            logger.setLevel(logging.INFO)
            fh = logging.FileHandler(file_name)
            fh.setLevel(logging.DEBUG)
            logger.addHandler(fh)

            mis.compute_infeasibility_explanation(m, opt, logger=logger)
            _check_output(file_name)


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
import pyomo.environ as pyo
from pyomo.contrib.mpc.data.get_cuid import get_indexed_cuid


class TestGetCUID(unittest.TestCase):
    def _make_model(self):
        m = pyo.ConcreteModel()
        m.time = pyo.Set(initialize=[0.0, 0.1, 0.2])
        m.space = pyo.Set(initialize=[1.0, 1.5, 2.0])
        m.comp = pyo.Set(initialize=["A", "B"])
        m.var = pyo.Var(m.time, m.comp, initialize=1.0)
        m.txc_var = pyo.Var(m.time, m.space, m.comp, initialize=2.0)

        @m.Block(m.time, m.space)
        def b(b, t, x):
            b.bvar1 = pyo.Var(initialize=3.0)
            b.bvar2 = pyo.Var(m.comp, initialize=3.0)

        return m

    def test_get_cuid(self):
        m = self._make_model()

        pred_cuid = pyo.ComponentUID(m.var[:, "A"])
        self.assertEqual(get_indexed_cuid(m.var[:, "A"]), pred_cuid)
        self.assertEqual(get_indexed_cuid(pyo.Reference(m.var[:, "A"])), pred_cuid)
        self.assertEqual(get_indexed_cuid("var[*,A]"), pred_cuid)
        self.assertEqual(get_indexed_cuid("var[*,'A']"), pred_cuid)
        self.assertEqual(get_indexed_cuid(m.var[0, "A"], sets=(m.time,)), pred_cuid)

    def test_get_cuid_twosets(self):
        m = self._make_model()

        pred_cuid = pyo.ComponentUID(m.b[:, :].bvar2["A"])
        self.assertEqual(get_indexed_cuid(m.b[:, :].bvar2["A"]), pred_cuid)
        self.assertEqual(
            get_indexed_cuid(pyo.Reference(m.b[:, :].bvar2["A"])), pred_cuid
        )
        self.assertEqual(get_indexed_cuid("b[*,*].bvar2[A]"), pred_cuid)
        self.assertEqual(
            get_indexed_cuid(m.b[0, 1].bvar2["A"], sets=(m.time, m.space)), pred_cuid
        )

    def test_get_cuid_dereference(self):
        m = self._make_model()
        m.ref = pyo.Reference(m.var[:, "A"])
        m.ref2 = pyo.Reference(m.ref)

        pred_cuid = pyo.ComponentUID(m.var[:, "A"])

        # ref is attached to the model, so by default we don't "dereference"
        self.assertNotEqual(get_indexed_cuid(m.ref), pred_cuid)
        self.assertEqual(get_indexed_cuid(m.ref), pyo.ComponentUID(m.ref[:]))

        # If we use dereference=True, we do dereference and get the CUID of
        # the underlying slice.
        self.assertEqual(get_indexed_cuid(m.ref, dereference=True), pred_cuid)

        # However, we only dereference once, so a reference-to-reference
        # does not reveal the underlying slice (of the original reference)
        self.assertNotEqual(get_indexed_cuid(m.ref2, dereference=True), pred_cuid)
        self.assertEqual(
            get_indexed_cuid(m.ref2, dereference=True), pyo.ComponentUID(m.ref[:])
        )

        # But if we use dereference=2, we allow two dereferences, and get
        # the original slice
        self.assertEqual(get_indexed_cuid(m.ref2, dereference=2), pred_cuid)

    def test_get_cuid_context(self):
        m = self._make_model()
        top = pyo.ConcreteModel()
        top.m = m

        pred_cuid = pyo.ComponentUID(m.var[:, "A"], context=m)
        self.assertEqual(get_indexed_cuid(m.var[:, "A"], context=m), pred_cuid)
        self.assertEqual(
            get_indexed_cuid(pyo.Reference(m.var[:, "A"]), context=m), pred_cuid
        )

        # This is what we would expect without a context arg
        full_cuid = pyo.ComponentUID(m.var[:, "A"])
        self.assertNotEqual(get_indexed_cuid("m.var[*,A]"), pred_cuid)
        self.assertEqual(get_indexed_cuid("m.var[*,A]"), full_cuid)

        msg = "Context is not allowed"
        with self.assertRaisesRegex(ValueError, msg):
            # Passing context with a string raises a reasonable
            # error from the CUID constructor
            get_indexed_cuid("m.var[*,A]", context=m)

        self.assertEqual(
            get_indexed_cuid(m.var[0, "A"], sets=(m.time,), context=m), pred_cuid
        )


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

"""Model Graph Generator Code"""

from pyomo.common.dependencies import networkx as nx
from pyomo.core import Constraint, Objective, Var, ComponentMap, SortComponents
from pyomo.core.expr import identify_variables
from pyomo.contrib.community_detection.event_log import _event_log


def generate_model_graph(
    model,
    type_of_graph,
    with_objective=True,
    weighted_graph=True,
    use_only_active_components=True,
):
    """
    Creates a networkX graph of nodes and edges based on a Pyomo optimization model

    This function takes in a Pyomo optimization model, then creates a graphical representation of the model with
    specific features of the graph determined by the user (see Parameters below).

    (This function is designed to be called by detect_communities, but can be used solely for the purpose of
    creating model graphs as well.)

    Parameters
    ----------
    model: Block
        a Pyomo model or block to be used for community detection
    type_of_graph: str
        a string that specifies the type of graph that is created from the model
        'constraint' creates a graph based on constraint nodes,
        'variable' creates a graph based on variable nodes,
        'bipartite' creates a graph based on constraint and variable nodes (bipartite graph).
    with_objective: bool, optional
        a Boolean argument that specifies whether or not the objective function is included in the graph; the
        default is True
    weighted_graph: bool, optional
        a Boolean argument that specifies whether a weighted or unweighted graph is to be created from the Pyomo
        model; the default is True (type_of_graph='bipartite' creates an unweighted graph regardless of this parameter)
    use_only_active_components: bool, optional
        a Boolean argument that specifies whether inactive constraints/objectives are included in the networkX graph

    Returns
    -------
    bipartite_model_graph/projected_model_graph: nx.Graph
        a NetworkX graph with nodes and edges based on the given Pyomo optimization model
    number_component_map: dict
        a dictionary that (deterministically) maps a number to a component in the model
    constraint_variable_map: dict
        a dictionary that maps a numbered constraint to a list of (numbered) variables that appear in the constraint
    """

    # Start off by making a bipartite graph (regardless of the value of type_of_graph), then if
    # type_of_graph = 'variable' or 'constraint', we will "collapse" this bipartite graph into a variable node
    # or constraint node graph

    # Initialize the data structure needed to keep track of edges in the graph (this graph will be made
    # without edge weights, because edge weights are not useful for this bipartite graph)
    edge_set = set()

    bipartite_model_graph = (
        nx.Graph()
    )  # Initialize NetworkX graph for the bipartite graph
    constraint_variable_map = (
        {}
    )  # Initialize map of the variables in constraint equations

    # Make a dict of all the components we need for the NetworkX graph (since we cannot use the components directly
    # in the NetworkX graph)
    if with_objective:
        component_number_map = ComponentMap(
            (component, number)
            for number, component in enumerate(
                model.component_data_objects(
                    ctype=(Constraint, Var, Objective),
                    active=use_only_active_components,
                    descend_into=True,
                    sort=SortComponents.deterministic,
                )
            )
        )
    else:
        component_number_map = ComponentMap(
            (component, number)
            for number, component in enumerate(
                model.component_data_objects(
                    ctype=(Constraint, Var),
                    active=use_only_active_components,
                    descend_into=True,
                    sort=SortComponents.deterministic,
                )
            )
        )

    # Create the reverse of component_number_map, which will be used in detect_communities to convert the node numbers
    # to their corresponding Pyomo modeling components
    number_component_map = dict(
        (number, comp) for comp, number in component_number_map.items()
    )

    # Add the components as nodes to the bipartite graph
    bipartite_model_graph.add_nodes_from(
        [node_number for node_number in range(len(component_number_map))]
    )

    # Loop through all constraints in the Pyomo model to determine what edges need to be created
    for model_constraint in model.component_data_objects(
        ctype=Constraint, active=use_only_active_components, descend_into=True
    ):
        numbered_constraint = component_number_map[model_constraint]

        # Create a list of the variable numbers that occur in the given constraint equation
        numbered_variables_in_constraint_equation = [
            component_number_map[constraint_variable]
            for constraint_variable in identify_variables(model_constraint.expr)
        ]

        # Update constraint_variable_map
        constraint_variable_map[numbered_constraint] = (
            numbered_variables_in_constraint_equation
        )

        # Create a list of all the edges that need to be created based on the variables in this constraint equation
        edges_between_nodes = [
            (numbered_constraint, numbered_variable_in_constraint)
            for numbered_variable_in_constraint in numbered_variables_in_constraint_equation
        ]

        # Update edge_set based on the determined edges between nodes
        edge_set.update(edges_between_nodes)

    # This if statement will be executed if the user chooses to include the objective function as a node in
    # the model graph
    if with_objective:
        # Use a loop to account for the possibility of multiple objective functions
        for objective_function in model.component_data_objects(
            ctype=Objective, active=use_only_active_components, descend_into=True
        ):
            numbered_objective = component_number_map[objective_function]

            # Create a list of the variable numbers that occur in the given objective function
            numbered_variables_in_objective = [
                component_number_map[objective_variable]
                for objective_variable in identify_variables(objective_function)
            ]

            # Update constraint_variable_map
            constraint_variable_map[numbered_objective] = (
                numbered_variables_in_objective
            )

            # Create a list of all the edges that need to be created based on the variables in the objective function
            edges_between_nodes = [
                (numbered_objective, numbered_variable_in_objective)
                for numbered_variable_in_objective in numbered_variables_in_objective
            ]

            # Update edge_set based on the determined edges between nodes
            edge_set.update(edges_between_nodes)

    # Add edges to bipartite_model_graph (the order in which edges are added can affect community detection, so
    # sorting prevents any unpredictable changes)
    bipartite_model_graph.add_edges_from(sorted(edge_set))

    if (
        type_of_graph == 'bipartite'
    ):  # This is the case where the user wants a bipartite graph, which we made above
        # Log important information with the following logger function
        _event_log(
            model,
            bipartite_model_graph,
            set(constraint_variable_map),
            type_of_graph,
            with_objective,
        )

        # Return the bipartite NetworkX graph, the dictionary of node numbers mapped to their respective Pyomo
        # components, and the map of constraints to the variables they contain
        return bipartite_model_graph, number_component_map, constraint_variable_map

    # At this point of the code, we will create the projected version of the bipartite
    # model graph (based on the specific value of type_of_graph)

    constraint_nodes = set(constraint_variable_map)
    if type_of_graph == 'constraint':
        graph_nodes = constraint_nodes
    else:
        variable_nodes = set(number_component_map) - constraint_nodes
        graph_nodes = variable_nodes

    try:
        if weighted_graph:
            projected_model_graph = nx.bipartite.weighted_projected_graph(
                bipartite_model_graph, graph_nodes
            )
        else:
            projected_model_graph = nx.bipartite.projected_graph(
                bipartite_model_graph, graph_nodes
            )
    except nx.exception.NetworkXAlgorithmError:
        # See Pyomo #2413: networkx now raises exceptions for invalid
        # projections.  This restores the (probably invalid) previous
        # behavior in community_detection
        projected_model_graph = nx.Graph()

    # Log important information with the following logger function
    _event_log(
        model,
        projected_model_graph,
        set(constraint_variable_map),
        type_of_graph,
        with_objective,
    )

    # Return the projected NetworkX graph, the dictionary of node numbers mapped to their respective Pyomo
    # components, and the map of constraints to the variables they contain
    return projected_model_graph, number_component_map, constraint_variable_map



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest
from pyomo.core.expr.symbol_map import SymbolMap
from pyomo.core.kernel.variable import variable
from pyomo.environ import ConcreteModel, Var


class TestSymbolMap(unittest.TestCase):
    def test_no_labeler(self):
        s = SymbolMap()
        v = variable()
        self.assertEqual(str(v), s.getSymbol(v))

        s = SymbolMap()
        m = ConcreteModel()
        m.x = Var()
        self.assertEqual('x', s.createSymbol(m.x))

        s = SymbolMap()
        m.y = Var([1, 2, 3])
        s.createSymbols(m.y.values())
        self.assertEqual(s.bySymbol, {'y[1]': m.y[1], 'y[2]': m.y[2], 'y[3]': m.y[3]})
        self.assertEqual(
            s.byObject, {id(m.y[1]): 'y[1]', id(m.y[2]): 'y[2]', id(m.y[3]): 'y[3]'}
        )

    def test_default_labeler(self):
        s = SymbolMap(lambda x: "_" + str(x))
        v = variable()
        self.assertEqual("_" + str(v), s.getSymbol(v))

        s = SymbolMap(lambda x: "_" + str(x))
        m = ConcreteModel()
        m.x = Var()
        self.assertEqual('_x', s.createSymbol(m.x))

        s = SymbolMap(lambda x: "_" + str(x))
        m.y = Var([1, 2, 3])
        s.createSymbols(m.y.values())
        self.assertEqual(
            s.bySymbol, {'_y[1]': m.y[1], '_y[2]': m.y[2], '_y[3]': m.y[3]}
        )
        self.assertEqual(
            s.byObject, {id(m.y[1]): '_y[1]', id(m.y[2]): '_y[2]', id(m.y[3]): '_y[3]'}
        )

    def test_custom_labeler(self):
        labeler = lambda x, y: "^" + str(x) + y

        s = SymbolMap(lambda x: "_" + str(x))
        v = variable()
        self.assertEqual("^" + str(v) + "~", s.getSymbol(v, labeler, "~"))

        s = SymbolMap(lambda x: "_" + str(x))
        m = ConcreteModel()
        m.x = Var()
        self.assertEqual('^x~', s.createSymbol(m.x, labeler, "~"))

        s = SymbolMap(lambda x: "_" + str(x))
        m.y = Var([1, 2, 3])
        s.createSymbols(m.y.values(), labeler, "~")
        self.assertEqual(
            s.bySymbol, {'^y[1]~': m.y[1], '^y[2]~': m.y[2], '^y[3]~': m.y[3]}
        )
        self.assertEqual(
            s.byObject,
            {id(m.y[1]): '^y[1]~', id(m.y[2]): '^y[2]~', id(m.y[3]): '^y[3]~'},
        )

    def test_existing_alias(self):
        s = SymbolMap()
        v1 = variable()
        s.alias(v1, "v")
        self.assertIs(s.aliases["v"], v1)
        v2 = variable()
        with self.assertRaises(RuntimeError):
            s.alias(v2, "v")
        s.alias(v1, "A")
        self.assertIs(s.aliases["v"], v1)
        self.assertIs(s.aliases["A"], v1)
        s.alias(v1, "A")
        self.assertIs(s.aliases["v"], v1)
        self.assertIs(s.aliases["A"], v1)

    def test_add_symbol(self):
        s = SymbolMap()
        m = ConcreteModel()
        m.x = Var()
        m.y = Var([1, 2, 3])
        s.addSymbol(m.x, 'x')
        self.assertEqual(s.bySymbol, {'x': m.x})
        self.assertEqual(s.byObject, {id(m.x): 'x'})
        with self.assertRaisesRegex(
            RuntimeError, r'SymbolMap.addSymbol\(\): duplicate symbol.'
        ):
            s.addSymbol(m.y, 'x')
        s = SymbolMap()
        s.addSymbol(m.x, 'x')
        with self.assertRaisesRegex(
            RuntimeError, r'SymbolMap.addSymbol\(\): duplicate object.'
        ):
            s.addSymbol(m.x, 'y')

    def test_add_symbols(self):
        m = ConcreteModel()
        m.x = Var()
        m.y = Var([1, 2, 3])

        s = SymbolMap()
        s.addSymbols((m.y[i], str(i)) for i in (1, 2, 3))
        self.assertEqual(s.bySymbol, {'1': m.y[1], '2': m.y[2], '3': m.y[3]})
        self.assertEqual(
            s.byObject, {id(m.y[1]): '1', id(m.y[2]): '2', id(m.y[3]): '3'}
        )
        with self.assertRaisesRegex(
            RuntimeError, r'SymbolMap.addSymbols\(\): duplicate symbol.'
        ):
            s.addSymbols([(m.y, '1')])

        s = SymbolMap()
        s.addSymbols((m.y[i], str(i)) for i in (1, 2, 3))
        with self.assertRaisesRegex(
            RuntimeError, r'SymbolMap.addSymbols\(\): duplicate object.'
        ):
            s.addSymbols([(m.y[2], 'x')])


if __name__ == "__main__":
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import pyomo.common.unittest as unittest

from pyomo.core import ConcreteModel, Var, Expression, Block, RangeSet, Any
import pyomo.core.expr as EXPR
from pyomo.core.base.expression import NamedExpressionData
from pyomo.gdp.util import (
    clone_without_expression_components,
    is_child_of,
    get_gdp_tree,
)
from pyomo.gdp import Disjunct, Disjunction


class TestGDPUtils(unittest.TestCase):
    def test_clone_without_expression_components(self):
        m = ConcreteModel()
        m.x = Var(initialize=5)
        m.y = Var(initialize=3)
        m.e = Expression(expr=m.x**2 + m.x - 1)

        base = m.x**2 + 1
        test = clone_without_expression_components(base, {})
        self.assertIs(base, test)
        self.assertEqual(base(), test())
        test = clone_without_expression_components(base, {id(m.x): m.y})
        self.assertEqual(3**2 + 1, test())

        base = m.e
        test = clone_without_expression_components(base, {})
        self.assertIsNot(base, test)
        self.assertEqual(base(), test())
        self.assertIsInstance(base, NamedExpressionData)
        self.assertIsInstance(test, EXPR.SumExpression)
        test = clone_without_expression_components(base, {id(m.x): m.y})
        self.assertEqual(3**2 + 3 - 1, test())

        base = m.e + m.x
        test = clone_without_expression_components(base, {})
        self.assertIsNot(base, test)
        self.assertEqual(base(), test())
        self.assertIsInstance(base, EXPR.SumExpression)
        self.assertIsInstance(test, EXPR.SumExpression)
        self.assertIsInstance(base.arg(0), NamedExpressionData)
        self.assertIsInstance(test.arg(0), EXPR.SumExpression)
        test = clone_without_expression_components(base, {id(m.x): m.y})
        self.assertEqual(3**2 + 3 - 1 + 3, test())

    def test_is_child_of(self):
        m = ConcreteModel()
        m.b = Block()
        m.b.b_indexed = Block([1, 2])
        m.b_parallel = Block()

        knownBlocks = {}
        self.assertFalse(
            is_child_of(parent=m.b, child=m.b_parallel, knownBlocks=knownBlocks)
        )
        self.assertEqual(len(knownBlocks), 2)
        self.assertFalse(knownBlocks.get(m))
        self.assertFalse(knownBlocks.get(m.b_parallel))
        self.assertTrue(
            is_child_of(parent=m.b, child=m.b.b_indexed[1], knownBlocks=knownBlocks)
        )
        self.assertEqual(len(knownBlocks), 4)
        self.assertFalse(knownBlocks.get(m))
        self.assertFalse(knownBlocks.get(m.b_parallel))
        self.assertTrue(knownBlocks.get(m.b.b_indexed[1]))
        self.assertTrue(knownBlocks.get(m.b.b_indexed))

    def test_gdp_tree(self):
        m = ConcreteModel()
        m.x = Var()
        m.block = Block()
        m.block.d1 = Disjunct()
        m.block.d1.dd1 = Disjunct()
        m.disj1 = Disjunct()
        m.block.disjunction = Disjunction(expr=[m.block.d1, m.disj1])
        m.block.d1.b = Block()
        m.block.d1.b.dd2 = Disjunct()
        m.block.d1.b.dd3 = Disjunct()
        m.block.d1.disjunction = Disjunction(
            expr=[m.block.d1.dd1, m.block.d1.b.dd2, m.block.d1.b.dd3]
        )
        m.block.d1.b.dd2.disjunction = Disjunction(expr=[[m.x >= 1], [m.x <= -1]])
        targets = (m,)
        knownBlocks = {}
        tree = get_gdp_tree(targets, m, knownBlocks)

        # check tree structure first
        vertices = tree.vertices
        self.assertEqual(len(vertices), 10)
        in_degrees = {
            m.block.d1: 1,
            m.block.disjunction: 0,
            m.disj1: 1,
            m.block.d1.disjunction: 1,
            m.block.d1.dd1: 1,
            m.block.d1.b.dd2: 1,
            m.block.d1.b.dd3: 1,
            m.block.d1.b.dd2.disjunction: 1,
            m.block.d1.b.dd2.disjunction.disjuncts[0]: 1,
            m.block.d1.b.dd2.disjunction.disjuncts[1]: 1,
        }
        for key, val in in_degrees.items():
            self.assertEqual(tree.in_degree(key), val)

        # This should be deterministic, so we can just check the order
        topo_sort = [
            m.block.disjunction,
            m.disj1,
            m.block.d1,
            m.block.d1.disjunction,
            m.block.d1.b.dd3,
            m.block.d1.b.dd2,
            m.block.d1.b.dd2.disjunction,
            m.block.d1.b.dd2.disjunction.disjuncts[1],
            m.block.d1.b.dd2.disjunction.disjuncts[0],
            m.block.d1.dd1,
        ]
        sort = tree.topological_sort()
        for i, node in enumerate(sort):
            self.assertIs(node, topo_sort[i])

    def add_indexed_disjunction(self, parent, m):
        parent.indexed = Disjunction(Any)
        parent.indexed[1] = [
            [sum(m.x[i] ** 2 for i in m.I) <= 1],
            [sum((3 - m.x[i]) ** 2 for i in m.I) <= 1],
        ]
        parent.indexed[0] = [
            [(m.x[1] - 1) ** 2 + m.x[2] ** 2 <= 1],
            [-((m.x[1] - 2) ** 2) - (m.x[2] - 3) ** 2 >= -1],
        ]

    def test_gdp_tree_indexed_disjunction(self):
        # This is to check that indexed components never actually appear as
        # nodes in the tree. We should only have DisjunctionDatas and
        # DisjunctDatas.
        m = ConcreteModel()
        m.I = RangeSet(1, 4)
        m.x = Var(m.I, bounds=(-2, 6))
        self.add_indexed_disjunction(m, m)

        targets = (m.indexed,)
        knownBlocks = {}
        tree = get_gdp_tree(targets, m, knownBlocks)

        vertices = tree.vertices
        self.assertEqual(len(vertices), 6)
        in_degrees = {
            m.indexed[0]: 0,
            m.indexed[1]: 0,
            m.indexed[0].disjuncts[0]: 1,
            m.indexed[0].disjuncts[1]: 1,
            m.indexed[1].disjuncts[0]: 1,
            m.indexed[1].disjuncts[1]: 1,
        }
        for key, val in in_degrees.items():
            self.assertEqual(tree.in_degree(key), val)

        topo_sort = [
            m.indexed[0],
            m.indexed[0].disjuncts[1],
            m.indexed[0].disjuncts[0],
            m.indexed[1],
            m.indexed[1].disjuncts[1],
            m.indexed[1].disjuncts[0],
        ]
        sort = tree.topological_sort()
        for i, node in enumerate(sort):
            self.assertIs(node, topo_sort[i])

    def test_gdp_tree_nested_indexed_disjunction(self):
        m = ConcreteModel()
        m.I = RangeSet(1, 4)
        m.x = Var(m.I, bounds=(-2, 6))
        m.disj1 = Disjunct()
        self.add_indexed_disjunction(m.disj1, m)
        m.disj2 = Disjunct()
        m.another_disjunction = Disjunction(expr=[m.disj1, m.disj2])

        # First, we still just give the indexed disjunction as a target, and
        # make sure that we don't pick up the parent Disjunct in the tree, since
        # it is not in targets.
        targets = (m.disj1.indexed,)
        knownBlocks = {}
        tree = get_gdp_tree(targets, m, knownBlocks)

        vertices = tree.vertices
        self.assertEqual(len(vertices), 6)
        in_degrees = {
            m.disj1.indexed[0]: 0,
            m.disj1.indexed[1]: 0,
            m.disj1.indexed[0].disjuncts[0]: 1,
            m.disj1.indexed[0].disjuncts[1]: 1,
            m.disj1.indexed[1].disjuncts[0]: 1,
            m.disj1.indexed[1].disjuncts[1]: 1,
        }
        for key, val in in_degrees.items():
            self.assertEqual(tree.in_degree(key), val)

        topo_sort = [
            m.disj1.indexed[0],
            m.disj1.indexed[0].disjuncts[1],
            m.disj1.indexed[0].disjuncts[0],
            m.disj1.indexed[1],
            m.disj1.indexed[1].disjuncts[1],
            m.disj1.indexed[1].disjuncts[0],
        ]
        sort = tree.topological_sort()
        for i, node in enumerate(sort):
            self.assertIs(node, topo_sort[i])

        # Now, let targets be everything and make sure that we get the correct
        # tree.
        targets = (m,)
        tree = get_gdp_tree(targets, m, knownBlocks)
        vertices = tree.vertices
        self.assertEqual(len(vertices), 9)
        # update that now the disjunctions have a parent
        in_degrees[m.disj1.indexed[0]] = 1
        in_degrees[m.disj1.indexed[1]] = 1
        # and add new nodes
        in_degrees[m.disj1] = 1
        in_degrees[m.disj2] = 1
        in_degrees[m.another_disjunction] = 0
        for key, val in in_degrees.items():
            self.assertEqual(tree.in_degree(key), val)

        topo_sort = [
            m.another_disjunction,
            m.disj2,
            m.disj1,
            m.disj1.indexed[1],
            m.disj1.indexed[1].disjuncts[1],
            m.disj1.indexed[1].disjuncts[0],
            m.disj1.indexed[0],
            m.disj1.indexed[0].disjuncts[1],
            m.disj1.indexed[0].disjuncts[0],
        ]
        sort = tree.topological_sort()
        for i, node in enumerate(sort):
            self.assertIs(node, topo_sort[i])


if __name__ == '__main__':
    unittest.main()



# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________
# Note: the self.mcpp.* functions are all C-style functions implemented
# in the compiled MC++ wrapper library
# Note: argument to pow must be an integer


import ctypes
import logging
import os

from pyomo.common.fileutils import Library
from pyomo.core import value, Expression
from pyomo.core.base.block import SubclassOf
from pyomo.core.base.expression import NamedExpressionData
from pyomo.core.expr.numvalue import nonpyomo_leaf_types
from pyomo.core.expr.numeric_expr import (
    AbsExpression,
    LinearExpression,
    NegationExpression,
    NPV_AbsExpression,
    NPV_ExternalFunctionExpression,
    NPV_NegationExpression,
    NPV_PowExpression,
    NPV_ProductExpression,
    NPV_SumExpression,
    NPV_UnaryFunctionExpression,
    PowExpression,
    ProductExpression,
    SumExpression,
    UnaryFunctionExpression,
    NPV_DivisionExpression,
    DivisionExpression,
)
from pyomo.core.expr.visitor import StreamBasedExpressionVisitor, identify_variables
from pyomo.common.collections import ComponentMap

logger = logging.getLogger('pyomo.contrib.mcpp')

path = os.path.dirname(__file__)

__version__ = "19.11.12"


def mcpp_available():
    """True if the MC++ shared object file exists. False otherwise."""
    return Library('mcppInterface').path() is not None


NPV_expressions = (
    NPV_AbsExpression,
    NPV_ExternalFunctionExpression,
    NPV_NegationExpression,
    NPV_PowExpression,
    NPV_ProductExpression,
    NPV_SumExpression,
    NPV_UnaryFunctionExpression,
    NPV_DivisionExpression,
)


def _MCPP_lib():
    """A singleton interface to the MC++ library"""
    if _MCPP_lib._mcpp is not None:
        return _MCPP_lib._mcpp

    _MCPP_lib._mcpp = mcpp = ctypes.CDLL(Library('mcppInterface').path())

    # Version number
    mcpp.get_version.restype = ctypes.c_char_p

    mcpp.toString.argtypes = [ctypes.c_void_p]
    mcpp.toString.restype = ctypes.c_char_p

    mcpp.lower.argtypes = [ctypes.c_void_p]
    mcpp.lower.restype = ctypes.c_double

    mcpp.upper.argtypes = [ctypes.c_void_p]
    mcpp.upper.restype = ctypes.c_double

    mcpp.concave.argtypes = [ctypes.c_void_p]
    mcpp.concave.restype = ctypes.c_double

    mcpp.convex.argtypes = [ctypes.c_void_p]
    mcpp.convex.restype = ctypes.c_double

    mcpp.subcc.argtypes = [ctypes.c_void_p, ctypes.c_int]
    mcpp.subcc.restype = ctypes.c_double

    mcpp.subcv.argtypes = [ctypes.c_void_p, ctypes.c_int]
    mcpp.subcv.restype = ctypes.c_double

    # Create MC type variable
    mcpp.newVar.argtypes = [
        ctypes.c_double,
        ctypes.c_double,
        ctypes.c_double,
        ctypes.c_int,
        ctypes.c_int,
    ]
    mcpp.newVar.restype = ctypes.c_void_p

    # Create MC type constant
    mcpp.newConstant.argtypes = [ctypes.c_double]
    mcpp.newConstant.restype = ctypes.c_void_p

    # Multiply MC objects
    mcpp.multiply.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
    mcpp.multiply.restype = ctypes.c_void_p

    # Add MC objects
    mcpp.add.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
    mcpp.add.restype = ctypes.c_void_p

    # pow(x, y) functions
    # y is integer
    mcpp.power.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
    mcpp.power.restype = ctypes.c_void_p
    # y is fractional
    mcpp.powerf.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
    mcpp.powerf.restype = ctypes.c_void_p
    # y is an expression
    mcpp.powerx.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
    mcpp.powerx.restype = ctypes.c_void_p

    # sqrt function
    mcpp.mc_sqrt.argtypes = [ctypes.c_void_p]
    mcpp.mc_sqrt.restype = ctypes.c_void_p

    # - MC Variable
    mcpp.negation.argtypes = [ctypes.c_void_p]
    mcpp.negation.restype = ctypes.c_void_p

    # fabs(MC Variable)
    mcpp.mc_abs.argtypes = [ctypes.c_void_p]
    mcpp.mc_abs.restype = ctypes.c_void_p

    # sin(MC Variable)
    mcpp.trigSin.argtypes = [ctypes.c_void_p]
    mcpp.trigSin.restype = ctypes.c_void_p

    # cos(MC Variable)
    mcpp.trigCos.argtypes = [ctypes.c_void_p]
    mcpp.trigCos.restype = ctypes.c_void_p

    # tan(MC Variable)
    mcpp.trigTan.argtypes = [ctypes.c_void_p]
    mcpp.trigTan.restype = ctypes.c_void_p

    # asin(MC Variable)
    mcpp.atrigSin.argtypes = [ctypes.c_void_p]
    mcpp.atrigSin.restype = ctypes.c_void_p

    # acos(MC Variable)
    mcpp.atrigCos.argtypes = [ctypes.c_void_p]
    mcpp.atrigCos.restype = ctypes.c_void_p

    # atan(MC Variable)
    mcpp.atrigTan.argtypes = [ctypes.c_void_p]
    mcpp.atrigTan.restype = ctypes.c_void_p

    # exp(MC Variable)
    mcpp.exponential.argtypes = [ctypes.c_void_p]
    mcpp.exponential.restype = ctypes.c_void_p

    # log(MC Variable)
    mcpp.logarithm.argtypes = [ctypes.c_void_p]
    mcpp.logarithm.restype = ctypes.c_void_p

    # Releases object from memory (prevent memory leaks)
    mcpp.release.argtypes = [ctypes.c_void_p]

    # Unary function exception wrapper
    mcpp.try_unary_fcn.argtypes = [ctypes.c_void_p, ctypes.c_void_p]
    mcpp.try_unary_fcn.restype = ctypes.c_void_p

    # Binary function exception wrapper
    mcpp.try_binary_fcn.argtypes = [ctypes.c_void_p, ctypes.c_void_p, ctypes.c_void_p]
    mcpp.try_binary_fcn.restype = ctypes.c_void_p

    # Error message retrieval
    mcpp.get_last_exception_message.restype = ctypes.c_char_p

    return mcpp


# Initialize the singleton to None
_MCPP_lib._mcpp = None


class MCPP_Error(Exception):
    pass


class MCPP_visitor(StreamBasedExpressionVisitor):
    """Creates an MC++ expression from the corresponding Pyomo expression.

    This class walks a pyomo expression tree and builds up the corresponding
    expression of type McCormick.

    Note on memory management: The MCPP_visitor will return a pointer to
    an MC++ interval object that was dynamically allocated within the C
    interface.  It is the caller's responsibility to call
    `mcpp_lib.release()` on that object to prevent a memory leak

    """

    def __init__(self, expression, improved_var_bounds=None):
        super(MCPP_visitor, self).__init__()
        self.mcpp = _MCPP_lib()
        so_file_version = self.mcpp.get_version()
        so_file_version = so_file_version.decode("utf-8")
        if not so_file_version == __version__:
            raise MCPP_Error(
                "Shared object file version %s is out of date with MC++ interface version %s. "
                "Please rebuild the library." % (so_file_version, __version__)
            )
        self.missing_value_warnings = []
        self.expr = expression
        vars = list(identify_variables(expression, include_fixed=False))
        self.num_vars = len(vars)
        # Map expression variables to MC variables
        self.known_vars = ComponentMap()
        # Map expression variables to their index
        self.var_to_idx = ComponentMap()
        # Pre-register all variables
        inf = float('inf')
        for i, var in enumerate(vars):
            self.var_to_idx[var] = i
            # check if improved variable bound is provided
            if improved_var_bounds is not None:
                lb, ub = improved_var_bounds.get(var, (-inf, inf))
            else:
                lb, ub = -inf, inf
            self.known_vars[var] = self.register_var(var, lb, ub)
        self.refs = None

    def walk_expression(self):
        self.refs = set()
        ans = super(MCPP_visitor, self).walk_expression(self.expr)
        self.refs = None
        return ans

    def exitNode(self, node, data):
        if isinstance(node, ProductExpression):
            ans = self.mcpp.multiply(data[0], data[1])
        elif isinstance(node, SumExpression):
            ans = data[0]
            for arg in data[1:]:
                ans = self.mcpp.add(ans, arg)
        elif isinstance(node, PowExpression):
            if type(node.arg(1)) == int:
                ans = self.mcpp.try_binary_fcn(self.mcpp.power, data[0], data[1])
            elif type(node.arg(1)) == float:
                ans = self.mcpp.try_binary_fcn(self.mcpp.powerf, data[0], data[1])
            else:
                ans = self.mcpp.try_binary_fcn(self.mcpp.powerx, data[0], data[1])
        elif isinstance(node, DivisionExpression):
            ans = self.mcpp.try_binary_fcn(self.mcpp.divide, data[0], data[1])
        elif isinstance(node, NegationExpression):
            ans = self.mcpp.negation(data[0])
        elif isinstance(node, AbsExpression):
            ans = self.mcpp.try_unary_fcn(self.mcpp.mc_abs, data[0])
        elif isinstance(node, LinearExpression):
            raise NotImplementedError(
                'Quicksum has bugs that prevent proper usage of MC++.'
            )
            # ans = self.mcpp.newConstant(node.constant)
            # for coef, var in zip(node.linear_coefs, node.linear_vars):
            #     ans = self.mcpp.add(
            #         ans,
            #         self.mcpp.multiply(
            #             self.mcpp.newConstant(coef),
            #             self.register_num(var)))
        elif isinstance(node, UnaryFunctionExpression):
            if node.name == "exp":
                ans = self.mcpp.try_unary_fcn(self.mcpp.exponential, data[0])
            elif node.name == "log":
                ans = self.mcpp.try_unary_fcn(self.mcpp.logarithm, data[0])
            elif node.name == "sin":
                ans = self.mcpp.try_unary_fcn(self.mcpp.trigSin, data[0])
            elif node.name == "cos":
                ans = self.mcpp.try_unary_fcn(self.mcpp.trigCos, data[0])
            elif node.name == "tan":
                ans = self.mcpp.try_unary_fcn(self.mcpp.trigTan, data[0])
            elif node.name == "asin":
                ans = self.mcpp.try_unary_fcn(self.mcpp.atrigSin, data[0])
            elif node.name == "acos":
                ans = self.mcpp.try_unary_fcn(self.mcpp.atrigCos, data[0])
            elif node.name == "atan":
                ans = self.mcpp.try_unary_fcn(self.mcpp.atrigTan, data[0])
            elif node.name == "sqrt":
                ans = self.mcpp.try_unary_fcn(self.mcpp.mc_sqrt, data[0])
            else:
                raise NotImplementedError("Unknown unary function: %s" % (node.name,))
        elif isinstance(node, NPV_expressions):
            ans = self.mcpp.newConstant(value(data[0]))
        elif type(node) in nonpyomo_leaf_types:
            ans = self.mcpp.newConstant(node)
        elif not node.is_expression_type():
            ans = self.register_num(node)
        elif type(node) in SubclassOf(Expression) or isinstance(
            node, NamedExpressionData
        ):
            ans = data[0]
        else:
            raise RuntimeError("Unhandled expression type: %s" % (type(node)))

        if ans is None:
            msg = self.mcpp.get_last_exception_message()
            msg = msg.decode("utf-8")
            raise MCPP_Error(msg)

        return ans

    def beforeChild(self, node, child, child_idx):
        if type(child) in nonpyomo_leaf_types:
            # This means the child is POD
            # i.e., int, float, string
            return False, self.mcpp.newConstant(child)
        elif not child.is_expression_type():
            # node is either a Param, Var, or NumericConstant
            return False, self.register_num(child)
        else:
            # this is an expression node
            return True, None

    def acceptChildResult(self, node, data, child_result, child_idx):
        self.refs.add(child_result)
        data.append(child_result)
        return data

    def register_num(self, num):
        """Registers a new number: Param, Var, or NumericConstant."""
        if num.is_fixed():
            return self.mcpp.newConstant(value(num))
        else:
            return self.known_vars[num]

    def register_var(self, var, lb, ub):
        """Registers a new variable."""
        var_idx = self.var_to_idx[var]
        inf = float('inf')

        # Guard against errant None values in lb and ub
        lb = -inf if lb is None else lb
        ub = inf if ub is None else ub

        lb = max(var.lb if var.has_lb() else -inf, lb)
        ub = min(var.ub if var.has_ub() else inf, ub)
        var_val = value(var, exception=False)
        if lb == -inf:
            lb = -500000
            logger.warning(
                'Var %s missing lower bound. Assuming LB of %s' % (var.name, lb)
            )
        if ub == inf:
            ub = 500000
            logger.warning(
                'Var %s missing upper bound. Assuming UB of %s' % (var.name, ub)
            )
        if var_val is None:
            var_val = (lb + ub) / 2
            self.missing_value_warnings.append(
                'Var %s missing value. Assuming midpoint value of %s'
                % (var.name, var_val)
            )
        return self.mcpp.newVar(lb, var_val, ub, self.num_vars, var_idx)

    def finalizeResult(self, node_result):
        # Note, the node_result should NOT be in self.refs
        #    self.refs.remove(node_result)
        assert node_result not in self.refs
        for r in self.refs:
            self.mcpp.release(r)
        return node_result


class McCormick:
    """
    This class takes the constructed expression from MCPP_Visitor and
    allows for MC methods to be performed on pyomo expressions.

    __repn__(self): returns a display of an MC expression in the form:
    F: [lower interval : upper interval ] [convex underestimator :
    concave overestimator ] [ (convex subgradient) : (concave subgradient]

    lower(self): returns a float of the lower interval bound that is valid
    across the entire domain

    upper(self): returns a float of the upper interval bound that is valid
    across the entire domain

    concave(self): returns a float of the concave overestimator at the
    current value() of each variable.

    convex(self): returns a float of the convex underestimator at the
    current value() of each variable.

    ##Note: In order to describe the concave and convex relaxations over
    the entire domain, it is necessary to use changePoint() to repeat the
    calculation at different points.

    subcc(self): returns a ComponentMap() that maps the pyomo variables
    to the subgradients of the McCormick concave overestimators at the
    current value() of each variable.

    subcv(self): returns a ComponentMap() that maps the pyomo variables
    to the subgradients of the McCormick convex underestimators at the
    current value() of each variable.

    def changePoint(self, var, point): updates the current value() on the
    pyomo side and the current point on the MC++ side.
    """

    def __init__(self, expression, improved_var_bounds=None):
        # Guarantee that McCormick objects have mc_expr defined
        self.mc_expr = None

        self.mcpp = _MCPP_lib()
        self.pyomo_expr = expression
        self.visitor = MCPP_visitor(expression, improved_var_bounds)
        self.mc_expr = self.visitor.walk_expression()

    def __del__(self):
        if self.mc_expr is not None:
            self.mcpp.release(self.mc_expr)
            self.mc_expr = None

    def __repn__(self):
        repn = self.mcpp.toString(self.mc_expr)
        repn = repn.decode("utf-8")
        return repn

    def __str__(self):
        return self.__repn__()

    def lower(self):
        return self.mcpp.lower(self.mc_expr)

    def upper(self):
        return self.mcpp.upper(self.mc_expr)

    def concave(self):
        self.warn_if_var_missing_value()
        return self.mcpp.concave(self.mc_expr)

    def convex(self):
        self.warn_if_var_missing_value()
        return self.mcpp.convex(self.mc_expr)

    def subcc(self):
        self.warn_if_var_missing_value()
        ans = ComponentMap()
        for key in self.visitor.var_to_idx:
            i = self.visitor.var_to_idx[key]
            ans[key] = self.mcpp.subcc(self.mc_expr, i)
        return ans

    def subcv(self):
        self.warn_if_var_missing_value()
        ans = ComponentMap()
        for key in self.visitor.var_to_idx:
            i = self.visitor.var_to_idx[key]
            ans[key] = self.mcpp.subcv(self.mc_expr, i)
        return ans

    def changePoint(self, var, point):
        var.set_value(point)
        # WARNING: TODO: this has side effects.  If we do not use a
        # fresh MCPP_visitor, we get segfaults and different results.
        self.visitor = MCPP_visitor(self.visitor.expr)
        self.mcpp.release(self.mc_expr)
        self.mc_expr = self.visitor.walk_expression()

    def warn_if_var_missing_value(self):
        if self.visitor.missing_value_warnings:
            for message in self.visitor.missing_value_warnings:
                logger.warning(message)
            self.visitor.missing_value_warnings = []



# -*- coding: utf-8 -*-
# ____________________________________________________________________________________
#
# Pyomo: Python Optimization Modeling Objects
# Copyright (c) 2008-2026 National Technology and Engineering Solutions of Sandia, LLC
# Under the terms of Contract DE-NA0003525 with National Technology and Engineering
# Solutions of Sandia, LLC, the U.S. Government retains certain rights in this
# software.  This software is distributed under the 3-clause BSD License.
# ____________________________________________________________________________________

import gc
import itertools
import logging
import os
import platform
import time
import sys

from io import StringIO, BytesIO

from pyomo.common.errors import DeveloperError
from pyomo.common.log import LoggingIntercept, LogStream
from pyomo.common.tempfiles import TempfileManager
import pyomo.common.tee as tee
import pyomo.common.unittest as unittest


class timestamper:
    """A 'TextIO'-like object that records the time when data was written to
    the stream."""

    def __init__(self):
        self.buf = []
        self.error = ""

    def write(self, data):
        for line in data.splitlines():
            line = line.strip()
            if not line:
                continue
            self.buf.append((time.time(), float(line)))

    def writelines(self, data):
        for line in data:
            self.write(line.strip())

    def flush(self):
        pass

    def check(self, *bases):
        """Map the recorded times to {0, 1} based on the range of times
        recorded: anything in the first two-thirds of the range is mapped to
        0, and anything in the last third is mapped to 1.  This
        "discretizes" the times so that we can reliably compare to
        baselines.

        """

        n = list(itertools.chain(*self.buf))
        cutoff = min(n) + (max(n) - min(n)) * 2.0 / 3.0
        result = [tuple(0 if i < cutoff else 1 for i in _) for _ in self.buf]
        if result not in bases:
            base = ' or '.join(str(_) for _ in bases)
            self.error = f"result {result} != baseline {base}\nRaw timing: {self.buf}"
            return False
        return True


class TestTeeStream(unittest.TestCase):
    def setUp(self):
        self.reenable_gc = gc.isenabled()
        gc.disable()
        gc.collect()
        # Set a short switch interval so that the threading tests behave
        # as expected
        self.switchinterval = sys.getswitchinterval()
        sys.setswitchinterval(tee._poll_interval / 100)

    def tearDown(self):
        sys.setswitchinterval(self.switchinterval)
        if self.reenable_gc:
            gc.enable()
            gc.collect()

    def test_stdout(self):
        a = StringIO()
        b = StringIO()
        with tee.TeeStream(a, b) as t:
            t.STDOUT.write("Hello\n")
        self.assertEqual(a.getvalue(), "Hello\n")
        self.assertEqual(b.getvalue(), "Hello\n")

    def test_err_and_out_are_different(self):
        with tee.TeeStream() as t:
            out = t.STDOUT
            self.assertIs(out, t.STDOUT)
            err = t.STDERR
            self.assertIs(err, t.STDERR)
            self.assertIsNot(out, err)

    def test_signal_flush(self):
        a = StringIO()
        with tee.TeeStream(a) as t:
            out = t.STDOUT
            self.assertIs(type(out), tee._SignalFlush)
            out.write("out1\n")
            out.writelines(["out2\n", "out3\n"])
        self.assertEqual(a.getvalue(), "out1\nout2\nout3\n")
        with tee.TeeStream(a) as t:
            err = t.STDERR
            self.assertIs(type(err), tee._AutoFlush)
            err.write("err1\n")
            err.writelines(["err2\n", "err3\n"])
        self.assertEqual(a.getvalue(), "out1\nout2\nout3\nerr1\nerr2\nerr3\n")
        # Note: 'is not writable' appears to work for all platforms
        # except PyPy, where the exception is "readonly attribute
        # 'name'"
        with self.assertRaisesRegex(
            AttributeError, '.*(is not writable)|(readonly attribute)'
        ):
            tee.TeeStream().STDOUT.name = 'foo'

    @unittest.skipIf(
        not tee._peek_available,
        "Requires the _mergedReader, but _peek_available==False",
    )
    def test_merge_out_and_err(self):
        # Test that the STDERR/STDOUT streams are merged correctly
        # (i.e., STDOUT is line buffered and STDERR is not).  This merge
        # logic is only applicable when using the merged reader (i.e.,
        # _peek_available is True)
        a = StringIO()
        b = StringIO()
        # make sure this doesn't accidentally become a very long wait
        assert tee._poll_interval <= 0.1
        with tee.TeeStream(a, b) as t:
            # This is a slightly nondeterministic (on Windows), so a
            # flush() and short pause should help
            t.STDOUT.write("Hello\nWorld")
            # NOTE: do not flush: we will test flush in the next test
            # t.STDOUT.flush()
            time.sleep(tee._poll_interval * 100)
            t.STDERR.write("interrupting\ncow")
            t.STDERR.flush()
            # For determinism, it is important that the STDERR message
            # appears in the output stream before we start shutting down
            # the TeeStream (which will dump the OUT and ERR in an
            # arbitrary order)
            start_time = time.time()
            while 'cow' not in a.getvalue() and time.time() - start_time < 1:
                time.sleep(tee._poll_interval)
        acceptable_results = {
            "Hello\ninterrupting\ncowWorld",  # expected
            "interrupting\ncowHello\nWorld",  # Windows occasionally puts
            # all error before stdout
        }
        self.assertIn(a.getvalue(), acceptable_results)
        self.assertEqual(b.getvalue(), a.getvalue())

    def test_merge_out_and_err_flush(self):
        # Test that the STDERR/STDOUT streams are merged correctly
        # (i.e., STDOUT is line buffered and STDERR is not).  This merge
        # logic is only applicable when using the merged reader (i.e.,
        # _peek_available is True)
        a = StringIO()
        b = StringIO()
        # make sure this doesn't accidentally become a very long wait
        assert tee._poll_interval <= 0.1
        with tee.TeeStream(a, b) as t:
            # This is a slightly nondeterministic (on Windows), so a
            # flush() and short pause should help
            t.STDOUT.write("Hello\nWorld")
            t.STDOUT.flush()
            time.sleep(tee._poll_interval * 100)
            t.STDERR.write("interrupting\ncow")
            t.STDERR.flush()
            # For determinism, it is important that the STDERR message
            # appears in the output stream before we start shutting down
            # the TeeStream (which will dump the OUT and ERR in an
            # arbitrary order)
            start_time = time.time()
            while 'cow' not in a.getvalue() and time.time() - start_time < 1:
                time.sleep(tee._poll_interval)
        acceptable_results = {
            "Hello\nWorldinterrupting\ncow",  # expected
            "interrupting\ncowHello\nWorld",  # Windows occasionally puts
            # all error before stdout
        }
        self.assertIn(a.getvalue(), acceptable_results)
        self.assertEqual(b.getvalue(), a.getvalue())

    def test_merged_out_and_err_without_peek(self):
        a = StringIO()
        b = StringIO()
        try:
            _tmp, tee._peek_available = tee._peek_available, False
            with tee.TeeStream(a, b) as t:
                # Ensure both threads are running
                t.STDOUT
                t.STDERR
                # ERR should come out before OUT, but this is slightly
                # nondeterministic, so a short pause should help
                t.STDERR.write("Hello\n")
                t.STDERR.flush()
                time.sleep(tee._poll_interval * 2)
                t.STDOUT.write("World\n")
        finally:
            tee._peek_available = _tmp
        self.assertEqual(a.getvalue(), "Hello\nWorld\n")
        self.assertEqual(b.getvalue(), "Hello\nWorld\n")

    def test_binary_tee(self):
        a = BytesIO()
        b = BytesIO()
        with tee.TeeStream(a, b) as t:
            t.open('wb').write(b"Hello\n")
        self.assertEqual(a.getvalue(), b"Hello\n")
        self.assertEqual(b.getvalue(), b"Hello\n")

    def test_decoder_and_buffer_errors(self):
        ref = "Hello, ©"
        bytes_ref = ref.encode()
        log = StringIO()
        with LoggingIntercept(log):
            # Note: we must force the encoding for Windows
            with tee.TeeStream(encoding='utf-8') as t:
                os.write(t.STDOUT.fileno(), bytes_ref[:-1])
        self.assertEqual(
            log.getvalue(),
            "Stream handle closed with un-decoded characters in the decoder "
            "buffer that was not emitted to the output stream(s):\n"
            "\tb'\\xc2'\n",
        )

        out = StringIO()
        log = StringIO()
        with LoggingIntercept(log):
            with tee.TeeStream(out) as t:
                out.close()
                t.STDOUT.write("hi\n")
        _id = hex(id(out))
        self.assertRegex(
            log.getvalue(),
            f"Error writing to output stream <_?io.StringIO @ {_id}>:"
            r"\n.*\nOutput stream closed before all output was written to it.\n"
            r"The following was left in the output buffer:\n    'hi\\n'\n$",
        )

        # TeeStream expects stream-like objects
        out = logging.getLogger()
        log = StringIO()
        with LoggingIntercept(log):
            with tee.TeeStream(out) as t:
                t.STDOUT.write("hi\n")
        _id = hex(id(out))
        self.assertRegex(
            log.getvalue(),
            f"Error writing to output stream <logging.RootLogger @ {_id}>:"
            r"\n.*\nIs this a writeable TextIOBase object\?\n"
            r"The following was left in the output buffer:\n    'hi\\n'\n$",
        )

        # Catch partial writes
        class fake_stream:
            def write(self, data):
                return 1

        out = fake_stream()
        log = StringIO()
        with LoggingIntercept(log):
            with tee.TeeStream(out) as t:
                t.STDOUT.write("hi\n")
        _id = hex(id(out))
        self.assertRegex(
            log.getvalue(),
            f"Incomplete write to output stream <.*fake_stream @ {_id}>."
            r"\nThe following was left in the output buffer:\n    'i\\n'\n$",
        )

    def test_context_mismatch(self):
        with self.assertRaisesRegex(
            RuntimeError, "TeeStream: exiting a context that was not entered"
        ):
            with tee.TeeStream() as t:
                t.__exit__(None, None, None)

    def test_handle_prematurely_closed(self):
        # Close the TextIO object
        with LoggingIntercept() as LOG:
            with tee.TeeStream() as t:
                t.STDOUT.close()
        self.assertEqual(LOG.getvalue(), "")

        # Close the underlying file descriptor
        with LoggingIntercept() as LOG:
            with tee.TeeStream() as t:
                os.close(t.STDOUT.fileno())
        self.assertEqual(LOG.getvalue(), "")


class TestCapture(unittest.TestCase):
    def setUp(self):
        self.streams = sys.stdout, sys.stderr
        self.fd = [os.dup(stream.fileno()) for stream in self.streams]
        self.reenable_gc = gc.isenabled()
        gc.disable()
        gc.collect()
        # Set a short switch interval so that the threading tests behave
        # as expected
        self.switchinterval = sys.getswitchinterval()
        sys.setswitchinterval(tee._poll_interval / 100)

    def tearDown(self):
        sys.stdout, sys.stderr = self.streams
        os.dup2(self.fd[0], self.streams[0].fileno())
        os.dup2(self.fd[1], self.streams[1].fileno())
        sys.setswitchinterval(self.switchinterval)
        if self.reenable_gc:
            gc.enable()
            gc.collect()

    def test_capture_output(self):
        out = StringIO()
        with tee.capture_output(out) as OUT:
            print('Hello World')
        self.assertEqual(OUT.getvalue(), 'Hello World\n')

    def test_duplicate_capture_output(self):
        out = StringIO()
        capture = tee.capture_output(out)
        capture.setup()
        try:
            with self.assertRaisesRegex(
                RuntimeError, 'Duplicate call to capture_output.setup'
            ):
                capture.setup()
        finally:
            capture.reset()

    def test_reset_capture_output_twice(self):
        capture = tee.capture_output()
        with capture as OUT1:
            print("test1")
        capture.reset()
        capture.reset()
        with capture as OUT2:
            print("test2")
        self.assertEqual(OUT1.getvalue(), "test1\n")
        self.assertEqual(OUT2.getvalue(), "test2\n")

    def test_capture_output_logfile_string(self):
        with TempfileManager.new_context() as tempfile:
            logfile = tempfile.create_tempfile()
            self.assertTrue(isinstance(logfile, str))
            with tee.capture_output(logfile):
                print('HELLO WORLD')
            with open(logfile, 'r') as f:
                result = f.read()
            self.assertEqual('HELLO WORLD\n', result)

        # Check that an exception when entering the context will cleanly
        # unroll the context stack
        logfile = os.path.join('path', 'to', 'nonexisting', 'file.txt')
        T = tee.capture_output(logfile)
        with self.assertRaises(FileNotFoundError):
            T.__enter__()
        self.assertEqual(T.context_stack, [])

    def test_capture_to_logger(self):
        logger = logging.getLogger('_pyomo_no_logger')
        lstream = LogStream(logging.WARNING, logger)
        orig = logger.propagate, logger.handlers
        try:
            logger.propagate = False
            logger.handlers = []
            with LoggingIntercept(module='_pyomo_no_logger') as LOG:
                with tee.capture_output(lstream, capture_fd=False):
                    sys.stderr.write("hi!\n")
                    sys.stderr.flush()
            self.assertEqual(LOG.getvalue(), "hi!\n")

            # test that we handle the lastResort logger correctly
            _lastResort = logging.lastResort
            with tee.capture_output() as OUT:
                with tee.capture_output(lstream, capture_fd=False):
                    self.assertIsNot(_lastResort, logging.lastResort)
                    sys.stderr.write("hi?\n")
            self.assertEqual(OUT.getvalue(), "hi?\n")

            # test that we allow redirect-to-logger out
            with tee.capture_output() as OUT:
                logger.addHandler(logging.NullHandler())
                logger.addHandler(logging.StreamHandler(sys.stderr))
                with tee.capture_output(lstream, capture_fd=False):
                    sys.stderr.write("hi.\n")
            self.assertEqual(OUT.getvalue(), "hi.\n")
            logger.handlers.clear()

            # test a sub-logger
            lstream = LogStream(
                logging.WARNING, logging.getLogger('_pyomo_no_logger.foo')
            )
            with tee.capture_output() as OUT:
                logger.addHandler(logging.NullHandler())
                logger.addHandler(logging.StreamHandler(sys.stderr))
                with tee.capture_output(lstream, capture_fd=False):
                    sys.stderr.write("hi,\n")
            self.assertEqual(OUT.getvalue(), "hi,\n")
        finally:
            logger.propagate, logger.handlers = orig

    def test_capture_fd_to_logger(self):
        logger = logging.getLogger('_pyomo_no_logger')
        lstream = LogStream(logging.WARNING, logger)
        orig = logger.propagate, logger.handlers
        try:
            logger.propagate = False
            logger.handlers = []
            with LoggingIntercept(module='_pyomo_no_logger') as LOG:
                with tee.capture_output(lstream, capture_fd=True):
                    sys.stderr.write("hi!\n")
                    sys.stderr.flush()
            self.assertEqual(LOG.getvalue(), "hi!\n")

            # test that we handle the lastResort logger correctly
            _lastResort = logging.lastResort
            with tee.capture_output() as OUT:
                with tee.capture_output(lstream, capture_fd=True):
                    self.assertIsNot(_lastResort, logging.lastResort)
                    sys.stderr.write("hi?\n")
            self.assertEqual(OUT.getvalue(), "hi?\n")

            # test that we allow redirect-to-logger out
            with tee.capture_output() as OUT:
                logger.addHandler(logging.NullHandler())
                logger.addHandler(logging.StreamHandler(sys.stderr))
                with tee.capture_output(lstream, capture_fd=True):
                    sys.stderr.write("hi.\n")
            self.assertEqual(OUT.getvalue(), "hi.\n")
            logger.handlers.clear()

            # test a sub-logger
            lstream = LogStream(
                logging.WARNING, logging.getLogger('_pyomo_no_logger.foo')
            )
            with tee.capture_output() as OUT:
                logger.addHandler(logging.NullHandler())
                logger.addHandler(logging.StreamHandler(sys.stderr))
                with tee.capture_output(lstream, capture_fd=True):
                    sys.stderr.write("hi,\n")
            self.assertEqual(OUT.getvalue(), "hi,\n")
        finally:
            logger.propagate, logger.handlers = orig

    def test_capture_to_logger_adapter(self):
        class Adapter(logging.LoggerAdapter):
            def process(self, msg, kwargs):
                return '[%s] %s' % (self.extra['foo'], msg), kwargs

        logger = logging.getLogger('_pyomo_no_logger')
        adapter = Adapter(logger, {"foo": 42})
        lstream = LogStream(logging.WARNING, adapter)
        orig = logger.propagate, logger.handlers
        try:
            logger.propagate = False
            logger.handlers = []
            with LoggingIntercept(module='_pyomo_no_logger') as LOG:
                with tee.capture_output(lstream, capture_fd=False):
                    sys.stderr.write("hi!\n")
                    sys.stderr.flush()
            self.assertEqual(LOG.getvalue(), "[42] hi!\n")

            # test that we handle the lastResort logger correctly
            _lastResort = logging.lastResort
            with tee.capture_output() as OUT:
                with tee.capture_output(lstream, capture_fd=False):
                    self.assertIsNot(_lastResort, logging.lastResort)
                    sys.stderr.write("hi?\n")
            self.assertEqual(OUT.getvalue(), "[42] hi?\n")

            # test that we allow redirect-to-logger out
            with tee.capture_output() as OUT:
                logger.addHandler(logging.NullHandler())
                logger.addHandler(logging.StreamHandler(sys.stderr))
                with tee.capture_output(lstream, capture_fd=False):
                    sys.stderr.write("hi.\n")
            self.assertEqual(OUT.getvalue(), "[42] hi.\n")
            logger.handlers.clear()
        finally:
            logger.propagate, logger.handlers = orig

    def test_capture_fd_to_logger_adapter(self):
        class Adapter(logging.LoggerAdapter):
            def process(self, msg, kwargs):
                return '[%s] %s' % (self.extra['foo'], msg), kwargs

        logger = logging.getLogger('_pyomo_no_logger')
        adapter = Adapter(logger, {"foo": 42})
        lstream = LogStream(logging.WARNING, adapter)
        orig = logger.propagate, logger.handlers
        try:
            logger.propagate = False
            logger.handlers = []
            with LoggingIntercept(module='_pyomo_no_logger') as LOG:
                with tee.capture_output(lstream, capture_fd=True):
                    sys.stderr.write("hi!\n")
                    sys.stderr.flush()
            self.assertEqual(LOG.getvalue(), "[42] hi!\n")

            # test that we handle the lastResort logger correctly
            _lastResort = logging.lastResort
            with tee.capture_output() as OUT:
                with tee.capture_output(lstream, capture_fd=True):
                    self.assertIsNot(_lastResort, logging.lastResort)
                    sys.stderr.write("hi?\n")
            self.assertEqual(OUT.getvalue(), "[42] hi?\n")

            # test that we allow redirect-to-logger out
            with tee.capture_output() as OUT:
                logger.addHandler(logging.NullHandler())
                logger.addHandler(logging.StreamHandler(sys.stderr))
                with tee.capture_output(lstream, capture_fd=True):
                    sys.stderr.write("hi.\n")
            self.assertEqual(OUT.getvalue(), "[42] hi.\n")
            logger.handlers.clear()
        finally:
            logger.propagate, logger.handlers = orig

    def test_no_fileno_stdout(self):
        T = tee.capture_output()
        with T:
            self.assertEqual(len(T.context_stack), 2)
        T = tee.capture_output(capture_fd=True)
        # out & err point to something other than fd 1 and 2
        sys.stdout = os.fdopen(os.dup(1), 'w', closefd=True)
        sys.stderr = os.fdopen(os.dup(2), 'w', closefd=True)
        with sys.stdout, sys.stderr:
            with T:
                self.assertEqual(len(T.context_stack), 8)
        # out & err point to fd 1 and 2
        sys.stdout = os.fdopen(1, 'w', closefd=False)
        sys.stderr = os.fdopen(2, 'w', closefd=False)
        with sys.stdout, sys.stderr:
            with T:
                self.assertEqual(len(T.context_stack), 6)
        # out & err have no fileno
        sys.stdout = StringIO()
        sys.stderr = StringIO()
        with sys.stdout, sys.stderr:
            with T:
                self.assertEqual(len(T.context_stack), 6)

    def test_closed_stdout(self):
        with tee.capture_output() as T_outer:
            sys.stdout.close()
            with tee.capture_output() as T_inner:
                print("test")
        self.assertEqual(T_outer.getvalue(), "")
        self.assertEqual(T_inner.getvalue(), "test\n")

    def test_capture_output_stack_error(self):
        OUT1 = StringIO()
        OUT2 = StringIO()
        old = (sys.stdout, sys.stderr)
        old_fd = os.dup(1), os.dup(2)
        try:
            a = tee.capture_output(OUT1)
            a.setup()
            b = tee.capture_output(OUT2)
            b.setup()
            with self.assertRaisesRegex(
                RuntimeError, 'Captured output .* does not match sys.stdout'
            ):
                a.reset()
        finally:
            # Clear b so that it doesn't call __exit__ and corrupt stdout/stderr
            b.tee = None
            os.dup2(old_fd[0], 1)
            os.dup2(old_fd[1], 2)
            sys.stdout, sys.stderr = old
            logging.getLogger('pyomo.common.tee').handlers.clear()

    def test_atomic_deadlock(self):
        save_poll = tee._poll_timeout_deadlock
        tee._poll_timeout_deadlock = 0.01

        co = tee.capture_output()
        try:
            tee.capture_output.startup_shutdown.acquire()
            with self.assertRaisesRegex(
                DeveloperError, "Deadlock starting capture_output"
            ):
                with tee.capture_output():
                    pass
            tee.capture_output.startup_shutdown.release()

            with self.assertRaisesRegex(
                DeveloperError, "Deadlock closing capture_output"
            ):
                with co:
                    tee.capture_output.startup_shutdown.acquire()
        finally:
            tee._poll_timeout_deadlock = save_poll
            if tee.capture_output.startup_shutdown.locked():
                tee.capture_output.startup_shutdown.release()
            co.reset()

    def test_capture_output_invalid_ostream(self):
        # Test that capture_output does not suppress errors from the tee
        # module
        _id = hex(id(15))
        with tee.capture_output(capture_fd=True) as OUT:
            with tee.capture_output(15):
                sys.stderr.write("hi\n")
        self.assertEqual(
            OUT.getvalue(),
            f"Error writing to output stream <builtins.int @ {_id}>:\n"
            "    AttributeError: 'int' object has no attribute 'write'\n"
            "Is this a writeable TextIOBase object?\n"
            "The following was left in the output buffer:\n    'hi\\n'\n",
        )

        with tee.capture_output(capture_fd=True) as OUT:
            with tee.capture_output(15, capture_fd=True):
                print("hi")
        self.assertEqual(
            OUT.getvalue(),
            f"Error writing to output stream <builtins.int @ {_id}>:\n"
            "    AttributeError: 'int' object has no attribute 'write'\n"
            "Is this a writeable TextIOBase object?\n"
            "The following was left in the output buffer:\n    'hi\\n'\n",
        )

    def test_exit_on_del(self):
        # This is a weird "feature", but because things like the pyomo
        # script will create and "enter" a capture_output object without
        # using a context manager, it is possible that the object can be
        # deleted without calling __exit__.  Check that the context
        # stack is correctly unwound
        T = tee.capture_output()
        T.__enter__()
        stack = T.context_stack
        self.assertGreater(len(stack), 0)
        del T
        # This is a bit tricky: for cpython, T should be immediately
        # deallocated (including calling __del__) through reference
        # counting.  pypy is trickier: because it lacks
        # reference-counting, it must rely on the GC.  We have seen
        # cases on GHA where a single call to gc.collect() was sometimes
        # insufficient to ensure that T was collected (but unable to
        # reproduce it locally).  We will try up to 4 times (1 more than
        # the number of generations in the GC)
        remaining_attempts = 4
        while len(stack) and remaining_attempts:
            gc.collect()
            time.sleep(((4 - remaining_attempts) / 4.0) ** 2)
            remaining_attempts -= 1
        try:
            self.assertEqual(len(stack), 0)
        except:
            # We still want to unwind the context managers if the test fails:
            while stack:
                try:
                    stack.pop().__exit__(None, None, None)
                except:
                    pass
            raise

    def test_deadlock(self):
        class MockStream:
            def write(self, data):
                # This test will kill the write before this sleep is
                # done, so a long sleep won't impact suite run time.
                time.sleep(1)

            def flush(self):
                pass

        _save = tee._poll_timeout, tee._poll_timeout_deadlock
        tee._poll_timeout = tee._poll_interval * 2**5  # 0.0032
        tee._poll_timeout_deadlock = tee._poll_interval * 2**7  # 0.0128
        try:
            with (
                LoggingIntercept() as LOG,
                self.assertRaisesRegex(RuntimeError, 'deadlock'),
            ):
                with tee.TeeStream(MockStream()) as t:
                    t.STDERR.write('*')
            self.assertEqual(
                'Significant delay observed waiting to join reader '
                'threads, possible output stream deadlock\n'
                'TeeStream: deadlock observed joining reader threads\n',
                LOG.getvalue(),
            )
        finally:
            tee._poll_timeout, tee._poll_timeout_deadlock = _save

        _save = tee._poll_timeout, tee._poll_timeout_deadlock
        tee._poll_timeout = tee._poll_interval * 2**5  # 0.0032
        tee._poll_timeout_deadlock = tee._poll_interval * 2**7  # 0.0128
        try:
            with (
                LoggingIntercept() as LOG,
                self.assertRaisesRegex(ValueError, 'testing'),
            ):
                with tee.TeeStream(MockStream()) as t:
                    t.STDERR.write('*')
                    raise ValueError('testing')
            self.assertEqual("", LOG.getvalue())
        finally:
            tee._poll_timeout, tee._poll_timeout_deadlock = _save


class BufferTester:
    def setUp(self):
        sys.stdout.flush()
        sys.stderr.flush()
        self.reenable_gc = gc.isenabled()
        gc.disable()
        gc.collect()
        # Set a short switch interval so that the threading tests behave
        # as expected
        self.switchinterval = sys.getswitchinterval()
        sys.setswitchinterval(tee._poll_interval)
        self.dt = 0.1

    def tearDown(self):
        sys.setswitchinterval(self.switchinterval)
        if self.reenable_gc:
            gc.enable()
            gc.collect()

    def test_buffered_stdout(self):
        # Test 1: short messages to STDOUT are buffered
        fd = self.capture_fd
        ts = timestamper()
        ts.write(f"{time.time()}\n")
        with tee.TeeStream(ts, ts) as t, tee.capture_output(t.STDOUT, capture_fd=fd):
            sys.stdout.write(f"{time.time()}\n")
            time.sleep(self.dt)
        ts.write(f"{time.time()}\n")
        baseline = [
            [(0, 0), (1, 0), (1, 0), (1, 1)],
            # TODO: [JDS] The stdout channel appears to sometimes be no
            # longer buffered.  I am not exactly sure why (my guess is
            # because the underlying pipe is not buffered), but as it is
            # generally not a problem to not buffer, we will put off
            # "fixing" it.
            [(0, 0), (0, 0), (0, 0), (1, 1)],
        ]
        if not ts.check(*baseline):
            self.fail(ts.error)

    def test_buffered_stdout_flush(self, retry=True):
        # Test 2: short messages to STDOUT that are flushed are flushed
        fd = self.capture_fd
        ts = timestamper()
        ts.write(f"{time.time()}\n")
        with tee.TeeStream(ts, ts) as t, tee.capture_output(t.STDOUT, capture_fd=fd):
            sys.stdout.write(f"{time.time()}\n")
            sys.stdout.flush()
            time.sleep(self.dt)
        ts.write(f"{time.time()}\n")
        if not ts.check([(0, 0), (0, 0), (0, 0), (1, 1)]):
            # FIXME: We don't know why, but this test will
            # intermittently fail.  For the moment, we will just wait a
            # little and give it a second chance with a longer delay.
            if retry:
                time.sleep(self.dt)
                self.dt *= 2.5
                self.test_buffered_stdout_flush(False)
            elif platform.python_implementation().lower().startswith('pypy'):
                # TODO: For some reason, some part of the flush logic is
                # not reliable under pypy.
                pass
            else:
                self.fail(ts.error)

    def test_buffered_stdout_long_message(self):
        # Test 3: long messages to STDOUT fill the buffer and are flushed
        fd = self.capture_fd
        ts = timestamper()
        ts.write(f"{time.time()}")
        with tee.TeeStream(ts, ts) as t, tee.capture_output(t.STDOUT, capture_fd=fd):
            # Note: bigger than the buffer we allocate on Windows.
            sys.stdout.write(f"{time.time()}" + ' ' * tee._pipe_buffersize + "\n")
            time.sleep(self.dt)
        ts.write(f"{time.time()}")
        if not ts.check([(0, 0), (0, 0), (0, 0), (1, 1)]):
            self.fail(ts.error)

    def test_buffered_stdout_embed_TeeStream(self):
        # Test 4: short messages captured directly to TeeStream are not
        # buffered.
        #
        # TODO: [JDS] I am not exactly sure why this is not buffered (my
        # guess is because the underlying pipe is not buffered), but as
        # it is generally not a problem to not buffer, we will put off
        # "fixing" it.
        fd = self.capture_fd
        ts = timestamper()
        ts.write(f"{time.time()}")
        with tee.capture_output(tee.TeeStream(ts, ts), capture_fd=fd):
            sys.stdout.write(f"{time.time()}\n")
            time.sleep(self.dt)
        ts.write(f"{time.time()}")
        if not ts.check([(0, 0), (0, 0), (0, 0), (1, 1)]):
            self.fail(ts.error)

    def test_buffered_stdout_flush_embed_TeeStream(self):
        # Test 5: short messages captured directly to TeeStream that are
        # flushed are flushed
        fd = self.capture_fd
        ts = timestamper()
        ts.write(f"{time.time()}")
        with tee.capture_output(tee.TeeStream(ts, ts), capture_fd=fd):
            sys.stdout.write(f"{time.time()}\n")
            sys.stdout.flush()
            time.sleep(self.dt)
        ts.write(f"{time.time()}")
        if not ts.check([(0, 0), (0, 0), (0, 0), (1, 1)]):
            self.fail(ts.error)

    def test_buffered_stdout_long_message_embed_TeeStream(self):
        # Test 6: long messages captured directly to TeeStream fill the
        # buffer and are flushed
        fd = self.capture_fd
        ts = timestamper()
        ts.write(f"{time.time()}")
        with tee.capture_output(tee.TeeStream(ts, ts), capture_fd=fd):
            sys.stdout.write(f"{time.time()}" + '    ' * 4096 + "\n")
            time.sleep(self.dt)
        ts.write(f"{time.time()}")
        if not ts.check([(0, 0), (0, 0), (0, 0), (1, 1)]):
            self.fail(ts.error)

    def test_buffered_stderr(self):
        # Test 1: short messages to STDERR are buffered, unless we are
        # capturing the underlying file descriptor, in which case they
        # are buffered.
        fd = self.capture_fd
        ts = timestamper()
        ts.write(f"{time.time()}")
        with tee.TeeStream(ts, ts) as t, tee.capture_output(t.STDOUT, capture_fd=fd):
            sys.stderr.write(f"{time.time()}\n")
            time.sleep(self.dt)
        ts.write(f"{time.time()}")
        if not ts.check([(0, 0), (0, 0), (0, 0), (1, 1)]):
            self.fail(ts.error)

    def test_buffered_stderr_flush(self):
        # Test 2: short messages to STDERR that are flushed are flushed
        fd = self.capture_fd
        ts = timestamper()
        ts.write(f"{time.time()}")
        with tee.TeeStream(ts, ts) as t, tee.capture_output(t.STDOUT, capture_fd=fd):
            sys.stderr.write(f"{time.time()}\n")
            sys.stderr.flush()
            time.sleep(self.dt)
        ts.write(f"{time.time()}")
        if not ts.check([(0, 0), (0, 0), (0, 0), (1, 1)]):
            self.fail(ts.error)

    def test_buffered_stderr_long_message(self):
        # Test 3: long messages to STDERR fill the buffer and are flushed
        fd = self.capture_fd
        ts = timestamper()
        ts.write(f"{time.time()}")
        with tee.TeeStream(ts, ts) as t, tee.capture_output(t.STDOUT, capture_fd=fd):
            sys.stderr.write(f"{time.time()}" + '  ' * 4096 + "\n")
            time.sleep(self.dt)
        ts.write(f"{time.time()}")
        if not ts.check([(0, 0), (0, 0), (0, 0), (1, 1)]):
            self.fail(ts.error)

    def test_buffered_stderr_embed_TeeStream(self):
        # Test 4: short messages captured directly to TeeStream are not
        # buffered, unless we are capturing the underlying file
        # descriptor, in which case they are buffered.
        fd = self.capture_fd
        ts = timestamper()
        ts.write(f"{time.time()}")
        with tee.capture_output(tee.TeeStream(ts, ts), capture_fd=fd):
            sys.stderr.write(f"{time.time()}\n")
            time.sleep(self.dt)
        ts.write(f"{time.time()}")
        if not ts.check([(0, 0), (0, 0), (0, 0), (1, 1)]):
            self.fail(ts.error)

    def test_buffered_stderr_flush_embed_TeeStream(self):
        # Test 5: short messages captured directly to TeeStream that are
        # flushed are flushed
        fd = self.capture_fd
        ts = timestamper()
        ts.write(f"{time.time()}")
        with tee.capture_output(tee.TeeStream(ts, ts), capture_fd=fd):
            sys.stderr.write(f"{time.time()}\n")
            sys.stderr.flush()
            time.sleep(self.dt)
        ts.write(f"{time.time()}")
        if not ts.check([(0, 0), (0, 0), (0, 0), (1, 1)]):
            self.fail(ts.error)

    def test_buffered_stderr_long_message_embed_TeeStream(self):
        # Test 6: long messages captured directly to TeeStream fill the
        # buffer and are flushed
        fd = self.capture_fd
        ts = timestamper()
        ts.write(f"{time.time()}")
        with tee.capture_output(tee.TeeStream(ts, ts), capture_fd=fd):
            sys.stderr.write(f"{time.time()}" + '  ' * 4096 + "\n")
            time.sleep(self.dt)
        ts.write(f"{time.time()}")
        if not ts.check([(0, 0), (0, 0), (0, 0), (1, 1)]):
            self.fail(ts.error)


class TestBuffering_noCapture(BufferTester, unittest.TestCase):
    capture_fd = False


class TestBuffering_capture(BufferTester, unittest.TestCase):
    capture_fd = True


class TestFileDescriptor(unittest.TestCase):
    def setUp(self):
        self.out = sys.stdout
        self.out_fd = os.dup(1)

    def tearDown(self):
        sys.stdout = self.out
        os.dup2(self.out_fd, 1)
        os.close(self.out_fd)

    def _generate_output(self, redirector):
        with redirector:
            sys.stdout.write("to_stdout_1\n")
            sys.stdout.flush()
            with os.fdopen(1, 'w', closefd=False) as F:
                F.write("to_fd1_1\n")
                F.flush()

        sys.stdout.write("to_stdout_2\n")
        sys.stdout.flush()
        with os.fdopen(1, 'w', closefd=False) as F:
            F.write("to_fd1_2\n")
            F.flush()

    def test_redirect_synchronize_stdout(self):
        r, w = os.pipe()
        os.dup2(w, 1)
        sys.stdout = os.fdopen(1, 'w', closefd=False)
        rd = tee.redirect_fd(synchronize=True)
        self._generate_output(rd)

        with os.fdopen(r, 'r') as FILE:
            os.close(w)
            os.close(1)
            self.assertEqual(FILE.read(), "to_stdout_2\nto_fd1_2\n")

    def test_redirect_no_synchronize_stdout(self):
        r, w = os.pipe()
        os.dup2(w, 1)
        sys.stdout = os.fdopen(1, 'w', closefd=False)
        rd = tee.redirect_fd(synchronize=False)
        self._generate_output(rd)

        with os.fdopen(r, 'r') as FILE:
            os.close(w)
            os.close(1)
            self.assertEqual(FILE.read(), "to_stdout_2\nto_fd1_2\n")

    # Pytest's default capture method causes failures for the following
    # two tests. This re-implementation of the capfd fixture allows
    # the capture to be disabled for those two test specifically.
    @unittest.pytest.fixture(autouse=True)
    def capfd(self, capfd):
        """
        Reimplementation needed for use in unittest.TestCase subclasses
        """
        self.capfd = capfd

    def test_redirect_synchronize_stdout_not_fd1(self):
        self.capfd.disabled()
        r, w = os.pipe()
        os.dup2(w, 1)
        rd = tee.redirect_fd(synchronize=True)
        self._generate_output(rd)

        with os.fdopen(r, 'r') as FILE:
            os.close(w)
            os.close(1)
            self.assertEqual(FILE.read(), "to_fd1_2\n")

    def test_redirect_no_synchronize_stdout_not_fd1(self):
        self.capfd.disabled()
        r, w = os.pipe()
        os.dup2(w, 1)
        rd = tee.redirect_fd(synchronize=False)
        self._generate_output(rd)

        with os.fdopen(r, 'r') as FILE:
            os.close(w)
            os.close(1)
            self.assertEqual(FILE.read(), "to_fd1_2\n")

    def test_redirect_synchronize_stringio(self):
        r, w = os.pipe()
        os.dup2(w, 1)
        try:
            sys.stdout, out = StringIO(), sys.stdout
            rd = tee.redirect_fd(synchronize=True)
            self._generate_output(rd)
        finally:
            sys.stdout, out = out, sys.stdout

        self.assertEqual(out.getvalue(), "to_stdout_2\n")
        with os.fdopen(r, 'r') as FILE:
            os.close(w)
            os.close(1)
            self.assertEqual(FILE.read(), "to_fd1_2\n")

    def test_redirect_no_synchronize_stringio(self):
        r, w = os.pipe()
        os.dup2(w, 1)
        try:
            sys.stdout, out = StringIO(), sys.stdout
            rd = tee.redirect_fd(synchronize=False)
            self._generate_output(rd)
        finally:
            sys.stdout, out = out, sys.stdout

        self.assertEqual(out.getvalue(), "to_stdout_1\nto_stdout_2\n")
        with os.fdopen(r, 'r') as FILE:
            os.close(w)
            os.close(1)
            self.assertEqual(FILE.read(), "to_fd1_2\n")

    def test_capture_output_fd(self):
        r, w = os.pipe()
        os.dup2(w, 1)
        sys.stdout = os.fdopen(1, 'w', closefd=False)
        with tee.capture_output(capture_fd=True) as OUT:
            sys.stdout.write("to_stdout_1\n")
            sys.stdout.flush()
            with os.fdopen(1, 'w', closefd=False) as F:
                F.write("to_fd1_1\n")
                F.flush()

        sys.stdout.write("to_stdout_2\n")
        sys.stdout.flush()
        with os.fdopen(1, 'w', closefd=False) as F:
            F.write("to_fd1_2\n")
            F.flush()

        self.assertEqual(OUT.getvalue(), "to_stdout_1\nto_fd1_1\n")
        with os.fdopen(r, 'r') as FILE:
            os.close(1)
            os.close(w)
            self.assertEqual(FILE.read(), "to_stdout_2\nto_fd1_2\n")

    def test_nested_capture_output(self):
        OUT2 = StringIO()
        r, w = os.pipe()
        os.dup2(w, 1)
        sys.stdout = stdout0 = os.fdopen(1, 'w', closefd=False)
        with tee.capture_output((sys.stdout, StringIO()), capture_fd=True) as (_, OUT1):
            stdout1 = sys.stdout
            self.assertIsNot(stdout0, stdout1)
            with tee.capture_output((sys.stdout, OUT2), capture_fd=True):
                stdout2 = sys.stdout
                self.assertIsNot(stdout1, stdout2)
                sys.stdout.write("to_stdout_1\n")
                sys.stdout.flush()
                with os.fdopen(1, 'w', closefd=False) as F:
                    F.write("to_fd1_1\n")
                    F.flush()

        sys.stdout.write("to_stdout_2\n")
        sys.stdout.flush()
        with os.fdopen(1, 'w', closefd=False) as F:
            F.write("to_fd1_2\n")
            F.flush()

        self.assertEqual(OUT1.getvalue(), "to_stdout_1\nto_fd1_1\n")
        self.assertEqual(OUT2.getvalue(), "to_stdout_1\nto_fd1_1\n")
        with os.fdopen(r, 'r') as FILE:
            os.close(1)
            os.close(w)
            self.assertEqual(
                FILE.read(), "to_stdout_1\nto_fd1_1\nto_stdout_2\nto_fd1_2\n"
            )


if __name__ == '__main__':
    unittest.main()
